import pytest
from nonebug import App
from nonebot import get_adapter
from nonebot.adapters.discord import Bot, Adapter
from arclet.alconna import Args, Option, Alconna, Subcommand, CommandMeta
from nonebot.adapters.discord.api.types import ApplicationCommandType, ApplicationCommandOptionType
from nonebot.adapters.discord.api.model import (
    Snowflake,
    ApplicationCommandData,
    ApplicationCommandInteractionDataOption,
)

from tests.fake import fake_discord_interaction_event


@pytest.mark.asyncio()
async def test_dc_ext(app: App):
    from nonebot_plugin_alconna import Match, on_alconna, load_from_path

    # from nonebot_plugin_alconna.adapters.discord import DiscordSlashExtension

    alc = Alconna(
        ["/"],
        "permission",
        Subcommand("add", Args["plugin", str]["priority?", int]),
        Option("remove", Args["plugin", str]["time?", int]),
        meta=CommandMeta(description="权限管理"),
    )
    matcher = on_alconna(alc)  # , extensions=[DiscordSlashExtension()])
    load_from_path("~adapters.discord")

    @matcher.assign("add")
    async def add(plugin: Match[str], priority: Match[int]):
        await matcher.finish(f"added {plugin.result} with {priority.result if priority.available else 0}")

    @matcher.assign("remove")
    async def remove(plugin: Match[str], time: Match[int]):
        await matcher.finish(f"removed {plugin.result} with {time.result if time.available else -1}")

    assert alc.parse("/permission add test 99").matched
    async with app.test_matcher(matcher) as ctx:
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter, bot_info=None, self_id="12345", auto_connect=False)
        event = fake_discord_interaction_event(
            data=ApplicationCommandData(
                id=Snowflake(1234567890),
                name="permission",
                type=ApplicationCommandType.CHAT_INPUT,
                options=[
                    ApplicationCommandInteractionDataOption(
                        name="add",
                        type=ApplicationCommandOptionType.SUB_COMMAND,
                        options=[
                            ApplicationCommandInteractionDataOption(
                                name="plugin", type=ApplicationCommandOptionType.STRING, value="test"
                            ),
                            ApplicationCommandInteractionDataOption(
                                name="priority", type=ApplicationCommandOptionType.INTEGER, value=99
                            ),
                        ],
                    )
                ],
            )
        )
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, "added test with 99")
        ctx.should_finished(matcher)

    assert alc.parse("/permission remove test 123").matched
    async with app.test_matcher(matcher) as ctx:
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter, bot_info=None, self_id="12345", auto_connect=False)
        event = fake_discord_interaction_event(
            data=ApplicationCommandData(
                id=Snowflake(1234567890),
                name="permission",
                type=ApplicationCommandType.CHAT_INPUT,
                options=[
                    ApplicationCommandInteractionDataOption(
                        name="remove",
                        type=ApplicationCommandOptionType.SUB_COMMAND,
                        options=[
                            ApplicationCommandInteractionDataOption(
                                name="plugin", type=ApplicationCommandOptionType.STRING, value="test"
                            ),
                            ApplicationCommandInteractionDataOption(
                                name="time", type=ApplicationCommandOptionType.INTEGER, value=123
                            ),
                        ],
                    )
                ],
            )
        )
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, "removed test with 123")
        ctx.should_finished(matcher)



from typing import TYPE_CHECKING, Any, Union

from tarina import lang
from nonebot.adapters import Bot, Event, Message

from ..export import Target, SupportAdapter, MessageExporter, SerializeFailed, export
from ..segment import At, Card, File, Text, AtAll, Audio, Emoji, Image, Reply, Voice, RefNode, Reference

if TYPE_CHECKING:
    from nonebot.adapters.mirai2.message import MessageSegment


class MiraiMessageExporter(MessageExporter["MessageSegment"]):
    @classmethod
    def get_adapter(cls) -> SupportAdapter:
        return SupportAdapter.mirai

    def get_message_type(self):
        from nonebot.adapters.mirai2.message import MessageChain

        return MessageChain

    def get_target(self, event: Event, bot: Union[Bot, None] = None) -> Target:
        from nonebot.adapters.mirai2.event import (
            BotMuteEvent,
            GroupMessage,
            FriendMessage,
            BotUnmuteEvent,
            MemberJoinEvent,
            MemberMuteEvent,
            GroupRecallEvent,
            BotJoinGroupEvent,
            BotLeaveEventKick,
            FriendRecallEvent,
            MemberUnmuteEvent,
            BotLeaveEventActive,
            MemberLeaveEventKick,
            MemberLeaveEventQuit,
            GroupStateChangeEvent,
            MemberStateChangeEvent,
        )

        if isinstance(event, FriendMessage):
            return Target(
                str(event.sender.id),
                private=True,
                platform=self.get_adapter(),
                self_id=bot.self_id if bot else None,
            )
        elif isinstance(event, GroupMessage):
            return Target(str(event.sender.group.id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        elif isinstance(event, (BotMuteEvent, BotUnmuteEvent)):
            return Target(
                str(event.operator.group.id),
                platform=self.get_adapter(),
                self_id=bot.self_id if bot else None,
            )
        elif isinstance(event, (MemberMuteEvent, MemberUnmuteEvent)):
            return Target(str(event.member.group.id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        elif isinstance(event, (BotJoinGroupEvent, BotLeaveEventActive, BotLeaveEventKick)):
            return Target(str(event.group.id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        elif isinstance(event, (MemberJoinEvent, MemberLeaveEventKick, MemberLeaveEventQuit)):
            return Target(str(event.member.group.id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        elif isinstance(event, MemberStateChangeEvent):
            return Target(str(event.member.group.id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        elif isinstance(event, GroupStateChangeEvent):
            return Target(str(event.group.id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        elif isinstance(event, FriendRecallEvent):
            return Target(
                str(event.author_id),
                private=True,
                platform=self.get_adapter(),
                self_id=bot.self_id if bot else None,
            )
        elif isinstance(event, GroupRecallEvent):
            return Target(str(event.group.id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        raise NotImplementedError

    def get_message_id(self, event: Event) -> str:
        from nonebot.adapters.mirai2.event import MessageEvent

        assert isinstance(event, MessageEvent)
        return str(event.source and event.source.id)

    @export
    async def text(self, seg: Text, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        return ms.plain(seg.text)

    @export
    async def at(self, seg: At, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        return ms.at(int(seg.target))

    @export
    async def at_all(self, seg: AtAll, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        return ms.at_all()

    @export
    async def emoji(self, seg: Emoji, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        return ms.face(int(seg.id), seg.name)

    @export
    async def media(self, seg: Union[Image, Voice, Audio], bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        name = seg.__class__.__name__.lower()
        method = {
            "image": ms.image,
            "voice": ms.voice,
            "audio": ms.voice,
        }[name]
        if seg.id:
            return method(seg.id)
        elif seg.url:
            return method(url=seg.url)
        elif seg.path:
            return method(path=str(seg.path))
        else:
            raise SerializeFailed(lang.require("nbp-uniseg", "invalid_segment").format(type=name, seg=seg))

    @export
    async def file(self, seg: File, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        return ms.file(seg.id, seg.name, 0)  # type: ignore

    @export
    async def card(self, seg: Card, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        return ms.xml(seg.raw) if seg.flag == "xml" else ms.app(seg.raw)

    @export
    async def reply(self, seg: Reply, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        return ms("$reply", {"message_id": seg.id})  # type: ignore

    @export
    async def reference(self, seg: Reference, bot: Bot) -> "MessageSegment":
        from nonebot.adapters.mirai2.message import MessageType

        ms = self.segment_class
        if not seg.content or not isinstance(seg.content, list):
            raise SerializeFailed(lang.require("nbp-uniseg", "invalid_segment").format(type="forward", seg=seg))
        nodes = []
        for node in seg.content:
            if isinstance(node, RefNode):
                if node.context:
                    nodes.append({"messageRef": {"messageId": node.id, "target": node.context}})
                else:
                    nodes.append({"messageId": node.id})
            else:
                content = self.get_message_type()([])
                if isinstance(node.content, str):
                    content.extend(self.get_message_type()(node.content))
                elif isinstance(node.content, list):
                    content.extend(await self.export(node.content, bot, True))  # type: ignore
                else:
                    content.extend(node.content)
                nodes.append(
                    {
                        "senderId": node.uid,
                        "senderName": node.name,
                        "time": int(node.time.timestamp()),
                        "messageChain": content,
                    }
                )
        return ms(MessageType.FORWARD, nodeList=nodes)

    async def send_to(self, target: Union[Target, Event], bot: Bot, message: Message):
        from nonebot.adapters.mirai2.bot import Bot as MiraiBot

        assert isinstance(bot, MiraiBot)
        if TYPE_CHECKING:
            assert isinstance(message, self.get_message_type())

        if message.has("Quote"):
            quote = message.pop(message.index("Quote")).data["id"]
        else:
            quote = None

        if isinstance(target, Event):
            target = self.get_target(target, bot)

        if target.private:
            return await bot.send_friend_message(target=int(target.id), message_chain=message, quote=quote)
        else:
            return await bot.send_group_message(target=int(target.id), message_chain=message, quote=quote)

    async def recall(self, mid: Any, bot: Bot, context: Union[Target, Event]):
        from nonebot.adapters.mirai2.bot import Bot as MiraiBot

        assert isinstance(bot, MiraiBot)
        await bot.recall(target=mid["messageId"])
        return

    def get_reply(self, mid: Any):
        return Reply(str(mid["messageId"]))



from datetime import datetime
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from nonebot.adapters.qq import MessageCreateEvent as MessageCreateEvent
    from nonebot.adapters.satori.event import MessageEvent as SatoriMessageEvent
    from nonebot.adapters.discord.event import ApplicationCommandInteractionEvent
    from nonebot.adapters.onebot.v11 import GroupMessageEvent as GroupMessageEventV11
    from nonebot.adapters.onebot.v11 import PrivateMessageEvent as PrivateMessageEventV11


def fake_group_message_event_v11(**field) -> "GroupMessageEventV11":
    from pydantic import create_model
    from nonebot.adapters.onebot.v11.event import Sender
    from nonebot.adapters.onebot.v11 import Message, GroupMessageEvent

    _Fake = create_model("_Fake", __base__=GroupMessageEvent)

    class FakeEvent(_Fake):
        time: int = 1000000
        self_id: int = 1
        post_type: Literal["message"] = "message"
        sub_type: str = "normal"
        user_id: int = 10
        message_type: Literal["group"] = "group"
        group_id: int = 10000
        message_id: int = 1
        message: Message = Message("test")
        raw_message: str = "test"
        font: int = 0
        sender: Sender = Sender(
            card="",
            nickname="test",
            role="member",
        )
        to_me: bool = False

        class Config:
            extra = "forbid"

    return FakeEvent(**field)


def fake_private_message_event_v11(**field) -> "PrivateMessageEventV11":
    from pydantic import create_model
    from nonebot.adapters.onebot.v11.event import Sender
    from nonebot.adapters.onebot.v11 import Message, PrivateMessageEvent

    _Fake = create_model("_Fake", __base__=PrivateMessageEvent)

    class FakeEvent(_Fake):
        time: int = 1000000
        self_id: int = 1
        post_type: Literal["message"] = "message"
        sub_type: str = "friend"
        user_id: int = 10
        message_type: Literal["private"] = "private"
        message_id: int = 1
        message: Message = Message("test")
        raw_message: str = "test"
        font: int = 0
        sender: Sender = Sender(nickname="test")
        to_me: bool = False

        class Config:
            extra = "forbid"

    return FakeEvent(**field)


def fake_discord_interaction_event(**field) -> "ApplicationCommandInteractionEvent":
    from pydantic import create_model
    from nonebot.adapters.discord.event import ApplicationCommandInteractionEvent

    _Fake = create_model("_Fake", __base__=ApplicationCommandInteractionEvent)
    field["type"] = 2
    field["id"] = 123456
    field["application_id"] = 123456789
    field["token"] = "sometoken"
    field["version"] = 1

    class FakeEvent(_Fake):
        pass

    return FakeEvent(**field)


def fake_message_event_satori(**field) -> "SatoriMessageEvent":
    from pydantic import create_model
    from nonebot.adapters.satori import Message
    from nonebot.adapters.satori.utils import Element
    from nonebot.adapters.satori.event import MessageEvent
    from nonebot.adapters.satori.models import User, Channel, ChannelType, InnerMessage

    _Fake = create_model("_Fake", __base__=MessageEvent)

    class FakeEvent(_Fake):
        id: int = 1
        type: str = "message-created"
        self_id: str = "123456789"
        platform: str = "satori"
        timestamp: datetime = datetime.fromtimestamp(1000000)
        channel: Channel = Channel(id="1", type=ChannelType.TEXT)
        user: User = User(id="1", name="test")
        message: InnerMessage = InnerMessage(id="1", content=[Element(type="text", attrs={"text": "test"})])
        to_me: bool = False

        class Config:
            extra = "allow"

    _message = field.pop("message", Message("test"))
    event = FakeEvent(**field)
    event._message = _message
    return event


def fake_message_event_guild(**field) -> "MessageCreateEvent":
    from pydantic import create_model
    from nonebot.adapters.qq.message import Message
    from nonebot.adapters.qq.models.guild import User
    from nonebot.adapters.qq.event import MessageCreateEvent

    _Fake = create_model("_Fake", __base__=MessageCreateEvent)

    class FakeEvent(_Fake):
        id: str = "1234"
        channel_id: str = "abcd"
        guild_id: str = "efgh"
        content: str = "test"
        author: User = User(id=field.pop("user_id", "123456789"), username=field.pop("username", "foobar"))
        _message = field.pop("message", Message("test"))

        class Config:
            extra = "forbid"

    return FakeEvent(**field)



from datetime import datetime, timedelta
from typing import Dict, List, Type, Union, Optional

from tarina import lang
from nonebot.rule import Rule
from nonebot.adapters import Event
from nonebot.adapters.discord import Bot
from nonebot.permission import Permission
from nonebot.dependencies import Dependent
from arclet.alconna import Args, Option, Alconna, Subcommand
from nonebot.adapters.discord.event import ApplicationCommandInteractionEvent
from nonebot.typing import T_State, T_Handler, T_RuleChecker, T_PermissionChecker
from nonebot.adapters.discord.commands.storage import _application_command_storage
from nonebot.adapters.discord.api.types import MessageFlag, InteractionCallbackType
from nepattern import ANY, FLOAT, NUMBER, INTEGER, MatchMode, BasePattern, UnionPattern
from nonebot.internal.matcher.matcher import current_bot, current_event, current_matcher
from nonebot.adapters.discord.commands.matcher import SlashCommandMatcher, ApplicationCommandConfig, on_slash_command
from nonebot.adapters.discord.message import (
    Message,
    EmbedSegment,
    MessageSegment,
    StickerSegment,
    ComponentSegment,
    ReferenceSegment,
    TimestampSegment,
    AttachmentSegment,
    CustomEmojiSegment,
    MentionRoleSegment,
    MentionUserSegment,
    MentionChannelSegment,
    MentionEveryoneSegment,
    parse_message,
)
from nonebot.adapters.discord.api import (
    MessageGet,
    RoleOption,
    UserOption,
    NumberOption,
    OptionChoice,
    StringOption,
    BooleanOption,
    ChannelOption,
    IntegerOption,
    SnowflakeType,
    AnyCommandOption,
    AttachmentOption,
    SubCommandOption,
    MentionableOption,
    InteractionResponse,
    SubCommandGroupOption,
    ApplicationCommandType,
    ApplicationCommandOptionType,
    ApplicationCommandInteractionDataOption,
)

from nonebot_plugin_alconna.uniseg import At
from nonebot_plugin_alconna.typings import SegmentPattern
from nonebot_plugin_alconna.uniseg import Image as UniImg
from nonebot_plugin_alconna import Extension, UniMessage, log
from nonebot_plugin_alconna.matcher import _M, AlconnaMatcher

Text = str
Embed = SegmentPattern("embed", EmbedSegment, MessageSegment.embed)
Sticker = SegmentPattern("sticker", StickerSegment, MessageSegment.sticker)
Component = SegmentPattern("component", ComponentSegment, MessageSegment.component)
Timestamp = SegmentPattern("timestamp", TimestampSegment, MessageSegment.timestamp)
Emoji = SegmentPattern("emoji", CustomEmojiSegment, MessageSegment.custom_emoji)
Image = SegmentPattern("attachment", AttachmentSegment, MessageSegment.attachment)
MentionUser = SegmentPattern("mention_user", MentionUserSegment, MessageSegment.mention_user)
MentionChannel = SegmentPattern("mention_channel", MentionChannelSegment, MessageSegment.mention_channel)
MentionRole = SegmentPattern("mention_role", MentionRoleSegment, MessageSegment.mention_role)
MentionEveryone = SegmentPattern("mention_everyone", MentionEveryoneSegment, MessageSegment.mention_everyone)
Reference = SegmentPattern("reference", ReferenceSegment, MessageSegment.reference)


MentionID = (
    UnionPattern[Union[str, MentionUserSegment]](
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=int,
                alias="MentionUser",
                addition_accepts=MentionUser,
                converter=lambda _, x: int(x.data["user_id"]),
            ),
            BasePattern(
                r"@(\d+)",
                mode=MatchMode.REGEX_CONVERT,
                origin=int,
                alias="@xxx",
                accepts=str,
                converter=lambda _, x: int(x[1]),
            ),
            INTEGER,
        ]
    )
    @ "mention_id"
)
"""
内置类型，允许传入提醒元素(Mention)或者'@xxxx'式样的字符串或者数字, 返回数字
"""


def _translate_args(args: Args) -> List[AnyCommandOption]:
    result = []
    for arg in args:
        if isinstance(arg.value, UnionPattern):
            if all(isinstance(x, str) for x in arg.value.base):
                result.append(
                    StringOption(
                        name=arg.name,
                        description=arg.notice or arg.name,
                        required=not arg.optional,
                        choices=[OptionChoice(name=x, value=x) for x in arg.value.base],  # type: ignore  # noqa: E501
                    )
                )
            elif all(isinstance(x, int) for x in arg.value.base):
                result.append(
                    IntegerOption(
                        name=arg.name,
                        description=arg.notice or arg.name,
                        required=not arg.optional,
                        choices=[
                            OptionChoice(name=str(x), value=x) for x in arg.value.base  # type: ignore  # noqa: E501
                        ],
                    )
                )
            elif all(isinstance(x, float) for x in arg.value.base):
                result.append(
                    NumberOption(
                        name=arg.name,
                        description=arg.notice or arg.name,
                        required=not arg.optional,
                        choices=[
                            OptionChoice(name=str(x), value=x) for x in arg.value.base  # type: ignore  # noqa: E501
                        ],
                    )
                )
            else:
                result.append(
                    StringOption(
                        name=arg.name,
                        description=arg.notice or arg.name,
                        required=not arg.optional,
                        choices=[OptionChoice(name=str(x), value=str(x)) for x in arg.value.base],
                    )
                )
            continue
        elif arg.value == ANY:
            result.append(
                StringOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value == INTEGER:
            result.append(
                IntegerOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value in (FLOAT, NUMBER):
            result.append(
                NumberOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value.origin is str:
            result.append(
                StringOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value.origin is bool:
            result.append(
                BooleanOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value == Image or arg.value.origin == UniImg:
            result.append(
                AttachmentOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value == MentionUser:
            result.append(
                UserOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value == MentionChannel:
            result.append(
                ChannelOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value == MentionRole:
            result.append(
                RoleOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        elif arg.value.origin is At:
            result.append(
                MentionableOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
        else:
            result.append(
                StringOption(
                    name=arg.name,
                    description=arg.notice or arg.name,
                    required=not arg.optional,
                )
            )
    return result


def _translate_options(opt: Union[Option, Subcommand]) -> Union[SubCommandGroupOption, SubCommandOption]:
    if isinstance(opt, Option):
        return SubCommandOption(
            name=opt.name,
            description=opt.help_text,
            options=_translate_args(opt.args),  # type: ignore
        )
    if not opt.args.empty and opt.options:
        log(
            "WARNING",
            lang.require("nbp-alc", "log.discord_ambiguous_subcommand").format(name=opt.name),
        )
    if not opt.args.empty:
        return SubCommandOption(
            name=opt.name, description=opt.help_text, options=_translate_args(opt.args)  # type: ignore
        )
    return SubCommandGroupOption(
        name=opt.name,
        description=opt.help_text,
        options=[_translate_options(sub) for sub in opt.options],  # type: ignore
    )


def translate(
    alc: Alconna,
    internal_id: Optional[str] = None,
    rule: Union[Rule, T_RuleChecker, None] = None,
    permission: Union[Permission, T_PermissionChecker, None] = None,
    *,
    name_localizations: Optional[Dict[str, str]] = None,
    description_localizations: Optional[Dict[str, str]] = None,
    default_member_permissions: Optional[str] = None,
    dm_permission: Optional[bool] = None,
    default_permission: Optional[bool] = None,
    nsfw: Optional[bool] = None,
    handlers: Optional[List[Union[T_Handler, Dependent]]] = None,
    temp: bool = False,
    expire_time: Union[datetime, timedelta, None] = None,
    priority: int = 1,
    block: bool = True,
    state: Optional[T_State] = None,
    _depth: int = 0,
) -> Type[SlashCommandMatcher]:
    if alc.prefixes != ["/"] or (not alc.prefixes and isinstance(alc.command, str) and not alc.command.startswith("/")):
        raise ValueError(lang.require("nbp-alc", "error.discord_prefix"))
    allow_opt = [
        opt
        for opt in alc.options
        if opt.name not in set().union(*alc.namespace_config.builtin_option_name.values())  # type: ignore
    ]
    if not alc.args.empty and allow_opt:
        log(
            "WARNING",
            lang.require("nbp-alc", "log.discord_ambiguous_command").format(cmd=alc.path),
        )
    if not (options := _translate_args(alc.args)):
        options = [_translate_options(opt) for opt in allow_opt]
    buffer = {
        "name": alc.command,
        "description": alc.meta.description,
        "options": options,
        **locals(),
    }
    buffer["_depth"] += 1
    buffer.pop("alc")
    return on_slash_command(**buffer)


class DiscordSlashExtension(Extension):
    application_command: ApplicationCommandConfig

    @property
    def priority(self) -> int:
        return 10

    @property
    def id(self) -> str:
        return "adapters.discord:DiscordSlashExtension"

    def __init__(
        self,
        internal_id: Optional[str] = None,
        name_localizations: Optional[Dict[str, str]] = None,
        description: Optional[str] = None,
        description_localizations: Optional[Dict[str, str]] = None,
        default_member_permissions: Optional[str] = None,
        dm_permission: Optional[bool] = None,
        default_permission: Optional[bool] = None,
        nsfw: Optional[bool] = None,
    ):
        self.internal_id = internal_id
        self.name_localizations = name_localizations
        self.description = description
        self.description_localizations = description_localizations
        self.default_member_permissions = default_member_permissions
        self.dm_permission = dm_permission
        self.default_permission = default_permission
        self.nsfw = nsfw
        super().__init__()
        self.using = False

    def post_init(self, alc: Alconna) -> None:
        if alc.prefixes != ["/"] or (
            not alc.prefixes and isinstance(alc.command, str) and not alc.command.startswith("/")
        ):
            return
        self.using = True
        allow_opt = [
            opt
            for opt in alc.options
            if opt.name not in set().union(*alc.namespace_config.builtin_option_name.values())  # type: ignore
        ]
        if not alc.args.empty and allow_opt:
            log(
                "WARNING",
                lang.require("nbp-alc", "log.discord_ambiguous_command").format(cmd=alc.path),
            )
        if not (options := _translate_args(alc.args)):
            options = [_translate_options(opt) for opt in allow_opt]
        config = ApplicationCommandConfig(
            type=ApplicationCommandType.CHAT_INPUT,
            name=alc.command,
            name_localizations=self.name_localizations,
            description=self.description or alc.meta.description,
            description_localizations=self.description_localizations,
            options=options,  # type: ignore
            default_member_permissions=self.default_member_permissions,
            dm_permission=self.dm_permission,
            default_permission=self.default_permission,
            nsfw=self.nsfw,
        )
        _application_command_storage[self.internal_id or config.name] = config
        self.application_command = config

    def validate(self, bot, event: Event) -> bool:
        if not self.using:
            return False
        if not isinstance(event, ApplicationCommandInteractionEvent):
            return False
        if event.data.name != self.application_command.name or event.data.type != self.application_command.type:
            return False
        if not event.data.guild_id and self.application_command.guild_ids is None:
            return True
        return bool(
            event.data.guild_id
            and self.application_command.guild_ids
            and event.data.guild_id in self.application_command.guild_ids
        )

    async def message_provider(self, event: Event, state: T_State, bot, use_origin: bool = False):
        if not isinstance(event, ApplicationCommandInteractionEvent):
            return
        data = event.data
        cmd = f"/{data.name}"

        def _handle_options(options: List[ApplicationCommandInteractionDataOption]):
            for opt in options:
                if opt.type in (
                    ApplicationCommandOptionType.SUB_COMMAND,
                    ApplicationCommandOptionType.SUB_COMMAND_GROUP,
                ):
                    yield f"{opt.name}"
                    if opt.options:
                        yield from _handle_options(opt.options)
                else:
                    yield f"{opt.value}"

        if data.options:
            cmd += " "
            cmd += " ".join(_handle_options(data.options))

        return Message(cmd.rstrip())

    @classmethod
    async def send_deferred_response(cls) -> None:
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        await bot.create_interaction_response(
            interaction_id=event.id,
            interaction_token=event.token,
            response=InteractionResponse(type=InteractionCallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE),
        )

    @classmethod
    async def send_response(cls, message: _M, fallback: bool = False):
        matcher: AlconnaMatcher = current_matcher.get()  # type: ignore
        return await matcher.send(message, fallback)

    @classmethod
    async def get_response(cls) -> MessageGet:
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        return await bot.get_origin_interaction_response(
            application_id=bot.application_id,
            interaction_token=event.token,
        )

    async def edit_response(
        self,
        message: _M,
        fallback: bool = False,
    ) -> None:
        matcher: AlconnaMatcher = current_matcher.get()  # type: ignore
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        _message = await matcher.executor.send_wrapper(bot, event, matcher.convert(message))
        if isinstance(_message, UniMessage):
            message_data = parse_message(await _message.export(bot, fallback))  # type: ignore
        else:
            message_data = parse_message(_message)  # type: ignore
        await bot.edit_origin_interaction_response(
            application_id=bot.application_id,
            interaction_token=event.token,
            **message_data,
        )

    @classmethod
    async def delete_response(cls) -> None:
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        await bot.delete_origin_interaction_response(
            application_id=bot.application_id,
            interaction_token=event.token,
        )

    async def send_followup_msg(
        self,
        message: _M,
        fallback: bool = False,
        flags: Optional[MessageFlag] = None,
    ) -> MessageGet:
        matcher: AlconnaMatcher = current_matcher.get()  # type: ignore
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        _message = await matcher.executor.send_wrapper(bot, event, matcher.convert(message))
        if isinstance(_message, UniMessage):
            message_data = parse_message(await _message.export(bot, fallback))  # type: ignore
        else:
            message_data = parse_message(_message)  # type: ignore
        if flags:
            message_data["flags"] = int(flags)
        return await bot.create_followup_message(
            application_id=bot.application_id,
            interaction_token=event.token,
            **message_data,
        )

    @classmethod
    async def get_followup_msg(cls, message_id: SnowflakeType):
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        return await bot.get_followup_message(
            application_id=bot.application_id,
            interaction_token=event.token,
            message_id=message_id,
        )

    async def edit_followup_msg(
        self,
        message_id: SnowflakeType,
        message: _M,
        fallback: bool = False,
    ) -> MessageGet:
        matcher: AlconnaMatcher = current_matcher.get()  # type: ignore
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        _message = await matcher.executor.send_wrapper(bot, event, matcher.convert(message))
        if isinstance(_message, UniMessage):
            message_data = parse_message(await _message.export(bot, fallback))  # type: ignore
        else:
            message_data = parse_message(_message)  # type: ignore
        return await bot.edit_followup_message(
            application_id=bot.application_id,
            interaction_token=event.token,
            message_id=message_id,
            **message_data,
        )

    @classmethod
    async def delete_followup_msg(cls, message_id: SnowflakeType) -> None:
        event = current_event.get()
        bot = current_bot.get()
        if not isinstance(event, ApplicationCommandInteractionEvent) or not isinstance(bot, Bot):
            raise ValueError("Invalid event or bot")
        await bot.delete_followup_message(
            application_id=bot.application_id,
            interaction_token=event.token,
            message_id=message_id,
        )


__extension__ = DiscordSlashExtension



from typing import Union

from nepattern.base import INTEGER
from nonebot.adapters.onebot.v11.message import MessageSegment
from nepattern import URL, MatchMode, BasePattern, UnionPattern

from nonebot_plugin_alconna.typings import SegmentPattern

Anonymous = SegmentPattern("anonymous", MessageSegment, MessageSegment.anonymous)
Text = str
At = SegmentPattern("at", MessageSegment, MessageSegment.at, additional=lambda x: x.data["qq"] != "all")
AtAll = SegmentPattern(
    "at", MessageSegment, lambda: MessageSegment.at("all"), additional=lambda x: x.data["qq"] == "all"
)
Contact = SegmentPattern("contact", MessageSegment, MessageSegment.contact)
Dice = SegmentPattern("dice", MessageSegment, MessageSegment.dice)
Face = SegmentPattern("face", MessageSegment, MessageSegment.face)
Forward = SegmentPattern("forward", MessageSegment, MessageSegment.forward)
Image = SegmentPattern("image", MessageSegment, MessageSegment.image)
Json = SegmentPattern("json", MessageSegment, MessageSegment.json)
Location = SegmentPattern("location", MessageSegment, MessageSegment.location)
Music = SegmentPattern("music", MessageSegment, MessageSegment.music)
Node = SegmentPattern("node", MessageSegment, MessageSegment.node)
Poke = SegmentPattern("poke", MessageSegment, MessageSegment.poke)
Record = SegmentPattern("record", MessageSegment, MessageSegment.record)
Reply = SegmentPattern("reply", MessageSegment, MessageSegment.reply)
RPS = SegmentPattern("rps", MessageSegment, MessageSegment.rps)
Shake = SegmentPattern("shake", MessageSegment, MessageSegment.shake)
Share = SegmentPattern("share", MessageSegment, MessageSegment.share)
Video = SegmentPattern("video", MessageSegment, MessageSegment.video)
Xml = SegmentPattern("xml", MessageSegment, MessageSegment.xml)


ImgOrUrl = (
    UnionPattern[Union[str, MessageSegment]](
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=str,
                converter=lambda _, x: x.data["url"],  # type: ignore
                alias="img",
                accepts=MessageSegment,
            ),
            URL,  # type: ignore
        ]
    )
    @ "img_url"
)
"""
内置类型, 允许传入图片元素(Image)或者链接(URL)，返回链接
"""

AtID = (
    UnionPattern[Union[str, MessageSegment]](
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=int,
                alias="At",
                addition_accepts=At,
                converter=lambda _, x: int(x.data["qq"]),
            ),
            BasePattern(
                r"@(\d+)",
                mode=MatchMode.REGEX_CONVERT,
                origin=int,
                alias="@xxx",
                accepts=str,
                converter=lambda _, x: int(x[1]),
            ),
            INTEGER,
        ]
    )
    @ "at_id"
)
"""
内置类型，允许传入提醒元素(At)或者'@xxxx'式样的字符串或者数字, 返回数字
"""



from typing import Union

from nepattern.base import URL, INTEGER
from nonebot.adapters import Message as BaseMessage
from nonebot.adapters.satori.message import Message
from nonebot.adapters.satori.message import At as _At
from nonebot.adapters.satori.message import Br as _Br
from nonebot.adapters.satori.message import Bold as _Bold
from nonebot.adapters.satori.message import Code as _Code
from nonebot.adapters.satori.message import File as _File
from nonebot.adapters.satori.message import Link as _Link
from nepattern import MatchMode, BasePattern, UnionPattern
from nonebot.adapters.satori.message import MessageSegment
from nonebot.adapters.satori.message import Audio as _Audio
from nonebot.adapters.satori.message import Image as _Image
from nonebot.adapters.satori.message import Sharp as _Sharp
from nonebot.adapters.satori.message import Video as _Video
from nonebot.adapters.satori.message import Author as _Author
from nonebot.adapters.satori.message import Italic as _Italic
from nonebot.adapters.satori.message import Spoiler as _Spoiler
from nonebot.adapters.satori.message import Paragraph as _Paragraph
from nonebot.adapters.satori.message import Subscript as _Subscript
from nonebot.adapters.satori.message import Underline as _Underline
from nonebot.adapters.satori.message import Superscript as _Superscript
from nonebot.adapters.satori.message import RenderMessage as _RenderMessage
from nonebot.adapters.satori.message import Strikethrough as _Strikethrough

from nonebot_plugin_alconna.argv import MessageArgv
from nonebot_plugin_alconna.typings import SegmentPattern, TextSegmentPattern

Text = str
At = SegmentPattern("at", _At, MessageSegment.at, lambda x: "id" in x.data)
AtRole = SegmentPattern("at", _At, MessageSegment.at_role, lambda x: "role" in x.data)
AtAll = SegmentPattern("at", _At, MessageSegment.at_all, lambda x: x.data.get("type") in ("all", "here"))
Sharp = SegmentPattern("sharp", _Sharp, MessageSegment.sharp)
Link = SegmentPattern("link", _Link, MessageSegment.link)
Image = SegmentPattern("img", _Image, MessageSegment.image)
Audio = SegmentPattern("audio", _Audio, MessageSegment.audio)
Video = SegmentPattern("video", _Video, MessageSegment.video)
File = SegmentPattern("file", _File, MessageSegment.file)
RenderMessage = SegmentPattern("message", _RenderMessage, MessageSegment.message)
Quote = SegmentPattern("quote", _RenderMessage, MessageSegment.quote)
Author = SegmentPattern("author", _Author, MessageSegment.author)

ImgOrUrl = (
    UnionPattern(
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=str,
                converter=lambda _, x: x.data["src"],
                alias="img",
                addition_accepts=Image,
            ),
            URL,
        ]
    )
    @ "img_url"
)
"""
内置类型, 允许传入图片元素(Image)或者链接(URL)，返回链接
"""

MentionID = (
    UnionPattern[Union[str, _At]](
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=int,
                alias="at",
                addition_accepts=At,
                converter=lambda _, x: "id" in x.data and int(x.data["id"]),
            ),
            BasePattern(
                r"@(\d+)",
                mode=MatchMode.REGEX_CONVERT,
                origin=int,
                alias="@xxx",
                accepts=str,
                converter=lambda _, x: int(x[1]),
            ),
            INTEGER,
        ]
    )
    @ "at_id"
)
"""
内置类型，允许传入提醒元素(Mention)或者'@xxxx'式样的字符串或者数字, 返回数字
"""


styles = {
    "record": {},
    "index": 0,
    "msg": "",
}


def builder(self: MessageArgv, data: BaseMessage):
    styles["msg"] = str(data)
    _index = 0
    for index, unit in enumerate(data):
        if not self.is_text(unit):
            self.raw_data.append(unit)
            self.ndata += 1
            continue
        if not unit.data["text"].strip():
            if not index or index == len(data) - 1:
                continue
            if not self.is_text(data[index - 1]) or not self.is_text(data[index + 1]):
                continue
        text = unit.data["text"]
        if unit.type == "text":
            self.raw_data.append(text)
            self.ndata += 1
            continue
        if self.raw_data and self.raw_data[-1].__class__ is str:
            self.raw_data[-1] = f"{self.raw_data[-1]}{text}"
        else:
            self.raw_data.append(text)
            self.ndata += 1
        start = styles["msg"].find(text, _index)
        _index = start + len(text)
        styles["record"][(start, _index)] = unit.type


def clean_style():
    styles["record"].clear()
    styles["index"] = 0


MessageArgv.custom_build(Message, builder=builder, cleanup=clean_style)


def locator(x: str, t: str):
    start = styles["msg"].find(x, styles["index"])
    if start == -1:
        return False
    styles["index"] = start + len(x)
    if (maybe := styles["record"].get((start, styles["index"]))) and maybe == t:
        return True
    return any(
        scale[0] <= start <= scale[1] and scale[0] <= styles["index"] <= scale[1] and styles["record"][scale] == t
        for scale in styles["record"]
    )


Br = TextSegmentPattern("br", _Br, MessageSegment.br, locator)
Paragraph = TextSegmentPattern("paragraph", _Paragraph, MessageSegment.paragraph, locator)
Bold = TextSegmentPattern("bold", _Bold, MessageSegment.bold, locator)
Italic = TextSegmentPattern("italic", _Italic, MessageSegment.italic, locator)
Underline = TextSegmentPattern("underline", _Underline, MessageSegment.underline, locator)
Strikethrough = TextSegmentPattern("strikethrough", _Strikethrough, MessageSegment.strikethrough, locator)
Spoiler = TextSegmentPattern("spoiler", _Spoiler, MessageSegment.spoiler, locator)
Code = TextSegmentPattern("code", _Code, MessageSegment.code, locator)
Superscript = TextSegmentPattern("superscript", _Superscript, MessageSegment.superscript, locator)
Subscript = TextSegmentPattern("subscript", _Subscript, MessageSegment.subscript, locator)



from typing import Dict

from ..export import MessageExporter
from .qq import QQMessageExporter  # noqa: F401
from .red import RedMessageExporter  # noqa: F401
from .ding import DingMessageExporter  # noqa: F401
from .dodo import DoDoMessageExporter  # noqa: F401
from .kook import KookMessageExporter  # noqa: F401
from .mirai import MiraiMessageExporter  # noqa: F401
from .feishu import FeishuMessageExporter  # noqa: F401
from .github import GithubMessageExporter  # noqa: F401
from .ntchat import NTChatMessageExporter  # noqa: F401
from .satori import SatoriMessageExporter  # noqa: F401
from .console import ConsoleMessageExporter  # noqa: F401
from .discord import DiscordMessageExporter  # noqa: F401
from .bilibili import BilibiliMessageExporter  # noqa: F401
from .onebot11 import Onebot11MessageExporter  # noqa: F401
from .onebot12 import Onebot12MessageExporter  # noqa: F401
from .telegram import TelegramMessageExporter  # noqa: F401
from .minecraft import MinecraftMessageExporter  # noqa: F401

MAPPING: Dict[str, MessageExporter] = {cls.get_adapter(): cls() for cls in MessageExporter.__subclasses__()}



from io import BytesIO
from typing import Any

from nonebot import require
from nonebot.internal.matcher import current_bot, current_event

from nonebot_plugin_alconna.uniseg.message import Receipt, UniMessage

try:
    require("nonebot_plugin_saa")
    from nonebot_plugin_saa.registries import MessageId
    from nonebot_plugin_saa import MessageFactory, extract_target
    from nonebot_plugin_saa.registries import Receipt as SaaReceipt
    from nonebot_plugin_saa.utils.exceptions import AdapterNotSupported
    from nonebot_plugin_saa.types.common_message_segment import Text, Image, Reply, Mention
except ImportError:
    raise ImportError("You need to install nonebot_plugin_saa to use this module.")


def convert(mf: MessageFactory) -> UniMessage:
    msg = UniMessage()
    for msf in mf:
        if isinstance(msf, Text):
            msg.text(msf.data["text"])
        elif isinstance(msf, Image):
            image = msf.data["image"]
            if isinstance(image, str):
                msg.image(url=image, name=msf.data["name"])
            elif isinstance(image, (bytes, BytesIO)):
                msg.image(raw=image, name=msf.data["name"])
            else:
                msg.image(path=image, name=msf.data["name"])
        elif isinstance(msf, Reply):
            msg.text(f"[回复:{msf.data['message_id'].dict()}]")
        elif isinstance(msf, Mention):
            user_id = msf.data["user_id"]
            if user_id == "all":
                msg.at_all()
            elif user_id == "here":
                msg.at_all(online=True)
            else:
                msg.at(user_id)
        else:
            msg.text(f"[未知:{msf}]")
    return msg


class UnisegMessageId(MessageId):
    adapter_name: str
    message_id: Any

    class Config:
        arbitrary_types_allowed = True


class UnisegReceipt(SaaReceipt):
    adapter_name: str
    data: Receipt

    async def revoke(self):
        return await self.data.recall()

    class Config:
        arbitrary_types_allowed = True

    @property
    def raw(self) -> Receipt:
        return self.data

    def extract_message_id(self):
        return UnisegMessageId(adapter_name=self.adapter_name, message_id=self.data.msg_ids[0])


async def send(self: MessageFactory, *, at_sender=False, reply=False) -> "SaaReceipt":
    """回复消息，仅能用在事件响应器中"""
    try:
        event = current_event.get()
        bot = current_bot.get()
    except LookupError as e:
        raise RuntimeError("send() 仅能在事件响应器中使用，主动发送消息请使用 send_to") from e
    try:
        target = extract_target(event, bot)
        return await self._do_send(bot, target, event, at_sender, reply)
    except (RuntimeError, AdapterNotSupported):
        receipt = await convert(self).send(target=event, bot=bot, at_sender=at_sender, reply_to=reply)
        return UnisegReceipt(bot_id=receipt.bot.self_id, adapter_name=receipt.exporter.get_adapter(), data=receipt)


MessageFactory.send = send



import pytest
import nonebot
from nonebug import NONEBOT_INIT_KWARGS

# 导入适配器
from nonebot.adapters.qq import Adapter as QQAdapter
from nonebot.adapters.satori import Adapter as SatoriAdapter
from nonebot.adapters.discord import Adapter as DiscordAdapter
from nonebot.adapters.onebot.v11 import Adapter as Onebot11Adapter


def pytest_configure(config: pytest.Config):
    config.stash[NONEBOT_INIT_KWARGS] = {
        "driver": "~fastapi+~httpx+~websockets",
        "log_level": "DEBUG",
    }


@pytest.fixture(scope="session", autouse=True)
def load_bot():
    # 加载适配器
    driver = nonebot.get_driver()
    driver.register_adapter(QQAdapter)
    driver.register_adapter(DiscordAdapter)
    driver.register_adapter(Onebot11Adapter)
    driver.register_adapter(SatoriAdapter)

    nonebot.require("nonebot_plugin_alconna")
    return None



from typing import Literal

import pytest
from nonebug import App
from nonebot import get_adapter
from nonebot.adapters.onebot.v11 import Bot, Adapter, Message

from tests.fake import fake_group_message_event_v11


@pytest.mark.asyncio()
async def test_funcommand(app: App):
    from nonebot_plugin_alconna import funcommand

    table = {
        "add": float.__add__,
        "sub": float.__sub__,
        "mul": float.__mul__,
        "div": float.__truediv__,
    }

    @funcommand()
    async def calc(op: Literal["add", "sub", "mul", "div"], a: float, b: float):
        """加法测试"""
        return f"{a} {op} {b} = {table[op](a, b)}"

    async with app.test_matcher(calc) as ctx:  # type: ignore
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter)
        event = fake_group_message_event_v11(message=Message("calc add 1.3 2.4"), user_id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, "1.3 add 2.4 = 3.7")



from typing import TYPE_CHECKING, Union

from nonebot.adapters import Bot, Event, Message

from ..segment import At, Text, Image
from ..export import Target, SupportAdapter, MessageExporter, export

if TYPE_CHECKING:
    from nonebot.adapters.github.message import MessageSegment  # type: ignore


class GithubMessageExporter(MessageExporter["MessageSegment"]):
    def get_message_type(self):
        from nonebot.adapters.github.message import Message  # type: ignore

        return Message

    @classmethod
    def get_adapter(cls) -> SupportAdapter:
        return SupportAdapter.github

    def get_message_id(self, event: Event) -> str:
        from nonebot.adapters.github.event import MessageEvent  # type: ignore

        assert isinstance(event, MessageEvent)
        return str(event.id)  # type: ignore

    @export
    async def text(self, seg: Text, bot: Bot) -> "MessageSegment":
        ms = self.segment_class

        return ms.text(seg.text)

    @export
    async def at(self, seg: At, bot: Bot) -> "MessageSegment":
        ms = self.segment_class

        return ms.text(f"@{seg.target}")

    @export
    async def image(self, seg: Image, bot: Bot) -> "MessageSegment":
        ms = self.segment_class

        assert seg.url, "github image segment must have url"
        return ms.text(f"![]({seg.url})")

    async def send_to(self, target: Union[Target, Event], bot: Bot, message: Message):
        raise NotImplementedError



from typing import TYPE_CHECKING, Any, Union, cast

from tarina import lang
from nonebot.adapters import Bot, Event, Message

from ..export import Target, SupportAdapter, MessageExporter, SerializeFailed, export
from ..segment import At, Card, File, Text, AtAll, Audio, Emoji, Image, Reply, Video, Voice

if TYPE_CHECKING:
    from nonebot.adapters.kaiheila.message import MessageSegment


class KookMessageExporter(MessageExporter["MessageSegment"]):
    def get_message_type(self):
        from nonebot.adapters.kaiheila.message import Message

        return Message

    @classmethod
    def get_adapter(cls) -> SupportAdapter:
        return SupportAdapter.kook

    def get_message_id(self, event: Event) -> str:
        from nonebot.adapters.kaiheila.event import MessageEvent

        assert isinstance(event, MessageEvent)
        return str(event.msg_id)

    def get_target(self, event: Event, bot: Union[Bot, None] = None) -> Target:
        if group_id := getattr(event, "group_id", None):
            return Target(str(group_id), platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        if user_id := getattr(event, "user_id", None):
            return Target(str(user_id), private=True, platform=self.get_adapter(), self_id=bot.self_id if bot else None)
        raise NotImplementedError

    @export
    async def text(self, seg: Text, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        if seg.style and "markdown" in seg.style:
            return ms.KMarkdown(seg.text)
        return ms.text(seg.text)

    @export
    async def at(self, seg: At, bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        if seg.flag == "role":
            return ms.mention_role(seg.target)
        elif seg.flag == "channel":
            return ms.KMarkdown(f"(chn){seg.target}(chn)")
        else:
            return ms.mention(seg.target)

    @export
    async def at_all(self, seg: AtAll, bot: Bot) -> "MessageSegment":
        ms = self.segment_class

        return ms.mention_here() if seg.here else ms.mention_all()

    @export
    async def emoji(self, seg: Emoji, bot: Bot) -> "MessageSegment":
        ms = self.segment_class

        if seg.name:
            return ms.KMarkdown(f"(emj){seg.name}(emj)[{seg.id}]")
        else:
            return ms.KMarkdown(f":{seg.id}:")

    @export
    async def media(self, seg: Union[Image, Voice, Video, Audio, File], bot: Bot) -> "MessageSegment":
        ms = self.segment_class
        if TYPE_CHECKING:
            from nonebot.adapters.kaiheila.bot import Bot as KBot

            assert isinstance(bot, KBot)
        name = seg.__class__.__name__.lower()

        if seg.id or seg.url:
            method = {
                "image": ms.image,
                "voice": ms.audio,
                "audio": ms.audio,
                "video": ms.video,
                "file": ms.file,
            }[name]
            return method(seg.id or seg.url)
        method = {
            "image": ms.local_image,
            "voice": ms.local_audio,
            "audio": ms.local_audio,
            "video": ms.local_video,
            "file": ms.local_file,
        }[name]
        if seg.raw:
            return method(seg.raw_bytes)
        elif seg.path:
            return method(seg.path)
        else:
            raise SerializeFailed(lang.require("nbp-uniseg", "invalid_segment").format(type=name, seg=seg))

    @export
    async def card(self, seg: Card, bot: Bot) -> "MessageSegment":
        ms = self.segment_class

        if seg.flag == "xml":
            raise SerializeFailed(
                lang.require("nbp-uniseg", "failed_segment").format(adapter="kook", seg=seg, target="Card")
            )
        return ms.Card(seg.raw)

    @export
    async def reply(self, seg: Reply, bot: Bot) -> "MessageSegment":
        ms = self.segment_class

        return ms.quote(seg.id)

    async def send_to(self, target: Union[Target, Event], bot: Bot, message: Message):
        from nonebot.adapters.kaiheila.bot import Bot as KBot

        assert isinstance(bot, KBot)
        if TYPE_CHECKING:
            assert isinstance(message, self.get_message_type())

        if isinstance(target, Event):
            target = self.get_target(target, bot)

        if target.private:
            return await bot.send_msg(message_type="private", user_id=target.id, message=message)
        else:
            return await bot.send_msg(message_type="channel", channel_id=target.id, message=message)

    async def recall(self, mid: Any, bot: Bot, context: Union[Target, Event]):
        from nonebot.adapters.kaiheila.bot import Bot as KBot
        from nonebot.adapters.kaiheila.event import PrivateMessageEvent
        from nonebot.adapters.kaiheila.api.model import MessageCreateReturn

        _mid: MessageCreateReturn = cast(MessageCreateReturn, mid)

        assert _mid.msg_id

        assert isinstance(bot, KBot)
        if isinstance(context, Target):
            if context.private:
                await bot.directMessage_delete(msg_id=_mid.msg_id)
            else:
                await bot.message_delete(msg_id=_mid.msg_id)
        elif isinstance(context, PrivateMessageEvent):
            await bot.directMessage_delete(msg_id=_mid.msg_id)
        else:
            await bot.message_delete(msg_id=_mid.msg_id)
        return

    async def edit(self, new: Message, mid: Any, bot: Bot, context: Union[Target, Event]):
        from nonebot.adapters.kaiheila.bot import Bot as KBot
        from nonebot.adapters.kaiheila.event import PrivateMessageEvent
        from nonebot.adapters.kaiheila.message import MessageSerializer
        from nonebot.adapters.kaiheila.api.model import MessageCreateReturn

        if TYPE_CHECKING:
            assert isinstance(new, self.get_message_type())
            assert isinstance(bot, KBot)

        _mid: MessageCreateReturn = cast(MessageCreateReturn, mid)
        if not _mid.msg_id:
            return
        data = await MessageSerializer(new).serialize(bot=bot)
        data.pop("type", None)
        data["msg_id"] = _mid.msg_id
        if isinstance(context, Target):
            if context.private:
                data.pop("quote", None)
                await bot.directMessage_update(**data)
            else:
                await bot.message_update(**data)
        elif isinstance(context, PrivateMessageEvent):
            data.pop("quote", None)
            await bot.directMessage_update(**data)
        else:
            await bot.message_update(**data)
        return

    def get_reply(self, mid: Any):
        from nonebot.adapters.kaiheila.api.model import MessageCreateReturn

        _mid: MessageCreateReturn = cast(MessageCreateReturn, mid)
        if not _mid.msg_id:
            raise NotImplementedError
        return Reply(_mid.msg_id)



from typing import Union

import pytest
from nonebug import App
from nonebot import get_adapter
from arclet.alconna import Args, Alconna
from nonebot.adapters.onebot.v11 import Bot, Adapter, Message, MessageSegment

from tests.fake import fake_group_message_event_v11


@pytest.mark.asyncio()
async def test_got_path(app: App):
    from nonebot_plugin_alconna import At, Match, UniMessage, on_alconna

    test_cmd = on_alconna(Alconna("test", Args["target?", Union[str, At]]))

    @test_cmd.handle()
    async def tt_h(target: Match[Union[str, At]]):
        if target.available:
            test_cmd.set_path_arg("target", target.result)

    @test_cmd.got_path("target", prompt="请输入目标")
    async def tt(target: Union[str, At]):
        await test_cmd.send(UniMessage(["ok\n", target]))

    async with app.test_matcher(test_cmd) as ctx:
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter)
        event = fake_group_message_event_v11(message=Message("test"), user_id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, "请输入目标", result=None)
        ctx.should_rejected(test_cmd)
        event = fake_group_message_event_v11(message=Message("1234"), user_id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, Message("ok\n1234"))

        event = fake_group_message_event_v11(message=Message("test"), user_id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, "请输入目标", result=None)
        ctx.should_rejected(test_cmd)
        event = fake_group_message_event_v11(message=Message(MessageSegment.at(1234)), user_id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, Message([MessageSegment.text("ok\n"), MessageSegment.at(1234)]))

        event = fake_group_message_event_v11(message=Message("test 1234"), user_id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, Message("ok\n1234"))



from nonebot.adapters import Message, MessageSegment


class FallbackSegment(MessageSegment["FallbackMessage"]):
    @classmethod
    def get_message_class(cls):
        return FallbackMessage

    def __str__(self) -> str:
        if self.type == "text":
            return self.data["text"]
        return f"[{self.type}][{self.data}]"

    def is_text(self) -> bool:
        return self.type == "text"

    @staticmethod
    def text(text: str) -> "FallbackSegment":
        return FallbackSegment("text", {"text": text})


class FallbackMessage(Message[FallbackSegment]):
    @classmethod
    def get_segment_class(cls):
        return FallbackSegment

    @staticmethod
    def _construct(msg: str):
        yield FallbackSegment.text(msg)



from nonebot.plugin import PluginMetadata

from .segment import At as At
from .segment import Card as Card
from .segment import File as File
from .segment import Text as Text
from .params import MsgId as MsgId
from .segment import AtAll as AtAll
from .segment import Audio as Audio
from .segment import Emoji as Emoji
from .segment import Image as Image
from .segment import Other as Other
from .segment import Reply as Reply
from .segment import Video as Video
from .segment import Voice as Voice
from .const import log  # noqa: F401
from .export import Target as Target
from .params import UniMsg as UniMsg
from .segment import Custom as Custom
from .tools import get_bot as get_bot
from .message import Receipt as Receipt
from .segment import RefNode as RefNode
from .segment import Segment as Segment
from .params import MessageId as MessageId
from .params import MsgTarget as MsgTarget
from .segment import Reference as Reference
from .message import UniMessage as UniMessage
from .segment import CustomNode as CustomNode
from .tools import image_fetch as image_fetch
from .params import MessageTarget as MessageTarget
from .export import SupportAdapter as SupportAdapter
from .export import SerializeFailed as SerializeFailed
from .segment import custom_register as custom_register
from .fallback import FallbackMessage as FallbackMessage
from .fallback import FallbackSegment as FallbackSegment
from .params import UniversalMessage as UniversalMessage
from .params import UniversalSegment as UniversalSegment

__version__ = "0.38.1"

__plugin_meta__ = PluginMetadata(
    name="Universal Segment 插件",
    description="提供抽象的消息元素，与跨平台接收/发送消息功能的库",
    usage="unimsg: UniMsg",
    homepage="https://github.com/nonebot/plugin-alconna/tree/master/src/nonebot_plugin_alconna/uniseg",
    type="library",
    supported_adapters=None,
    extra={
        "author": "RF-Tar-Railt",
        "priority": 1,
        "version": __version__,
    },
)


def patch_saa():
    from .utils import saa  # noqa: F401



from nonebot.adapters.minecraft.message import MessageSegment

from nonebot_plugin_alconna.typings import SegmentPattern

Text = str
Title = SegmentPattern("title", MessageSegment, MessageSegment.title)
ActionBar = SegmentPattern("actionbar", MessageSegment, MessageSegment.actionbar)



import pytest
from nonebug import App
from arclet.alconna import Args, Alconna
from nonebot import require, get_adapter
from nonebot.adapters.satori import Bot, Adapter, MessageSegment

from tests.fake import fake_message_event_satori


@pytest.mark.asyncio()
async def test_patch(app: App):
    require("nonebot_plugin_saa")
    from nonebot_plugin_saa import Text, Mention, MessageFactory

    from nonebot_plugin_alconna import At, patch_saa, on_alconna

    test_cmd = on_alconna(Alconna("test", Args["target", At]))

    @test_cmd.handle()
    async def tt_h(target: At):
        await MessageFactory(
            [
                Text("ok\n"),
                Mention(target.target),
            ]
        ).send()

    patch_saa()

    async with app.test_matcher(test_cmd) as ctx:
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter, platform="satori", info=None)
        msg = "test" + MessageSegment.at("234")
        event = fake_message_event_satori(message=msg, id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_api(
            "message_create",
            {
                "channel_id": "1",
                "content": 'ok\n<at id="234"/>',
            },
        )



import asyncio
import importlib
from typing import Set, Dict, List, Type, Tuple, Union, Literal, Optional, cast

from nonebot.typing import T_State
from tarina import lang, init_spec
from nonebot.matcher import Matcher
from nonebot.utils import escape_tag
from nonebot.params import EventMessage
from nonebot.plugin.on import on_message
from nonebot.internal.rule import Rule as Rule
from nonebot.adapters import Bot, Event, Message
from nonebot import get_driver, get_plugin_config
from arclet.alconna.exceptions import SpecialOptionTriggered
from arclet.alconna import Alconna, Arparma, CompSession, output_manager, command_manager

from .config import Config
from .adapters import MAPPING
from .uniseg import UniMessage
from .model import CompConfig, CommandResult
from .extension import Extension, ExtensionExecutor
from .consts import ALCONNA_RESULT, ALCONNA_EXTENSION, ALCONNA_EXEC_RESULT, log

_modules = set()


class AlconnaRule:
    """检查消息字符串是否能够通过此 Alconna 命令。

    参数:
        command: Alconna 命令
        skip_for_unmatch: 是否在命令不匹配时跳过该响应
        auto_send_output: 是否自动发送输出信息并跳过响应
        comp_config: 自动补全配置
        extensions: 需要加载的匹配扩展
        exclude_ext: 需要排除的匹配扩展
        use_origin: 是否使用未经 to_me 等处理过的消息
        use_cmd_start: 是否使用 nb 全局配置里的命令前缀
    """

    __slots__ = (
        "command",
        "skip",
        "auto_send",
        "comp_config",
        "use_origin",
        "executor",
        "_waiter",
        "_matchers",
        "_futures",
        "_interfaces",
        "_comp_help",
    )

    def __init__(
        self,
        command: Alconna,
        skip_for_unmatch: bool = True,
        auto_send_output: bool = False,
        comp_config: Optional[CompConfig] = None,
        extensions: Optional[List[Union[Type[Extension], Extension]]] = None,
        exclude_ext: Optional[List[Union[Type[Extension], str]]] = None,
        use_origin: bool = False,
        use_cmd_start: bool = False,
        use_cmd_sep: bool = False,
        _aliases: Optional[Union[Set[str], Tuple[str, ...]]] = None,
    ):
        self.comp_config = comp_config
        self.use_origin = use_origin
        try:
            global_config = get_driver().config
            config = get_plugin_config(Config)
            self.auto_send = auto_send_output or config.alconna_auto_send_output
            if (
                not command.prefixes
                and (use_cmd_start or config.alconna_use_command_start)
                and global_config.command_start
            ):
                command_manager.delete(command)
                command.prefixes = list(global_config.command_start)
                command._hash = command._calc_hash()
                command_manager.register(command)
            if (use_cmd_sep or config.alconna_use_command_sep) and global_config.command_sep:
                command.separators = tuple(global_config.command_sep)
                command_manager.resolve(command).separators = tuple(global_config.command_sep)
            if config.alconna_auto_completion and not self.comp_config:
                self.comp_config = cast(CompConfig, {})
            if config.alconna_context_style:
                self.command.meta.context_style = config.alconna_context_style
                command_manager.resolve(command).context_style = config.alconna_context_style
            self.use_origin = use_origin or config.alconna_use_origin
        except ValueError:
            self.auto_send = auto_send_output
        self.command = command
        if _aliases:
            for alias in _aliases:
                command.shortcut(alias, prefix=True)
        self.skip = skip_for_unmatch
        self.executor = ExtensionExecutor(self, extensions, exclude_ext)
        self.executor.post_init()
        self._futures: Dict[str, Dict[str, asyncio.Future]] = {}
        self._matchers: Dict[str, Type[Matcher]] = {}
        self._interfaces: Dict[str, CompSession] = {}

        self._comp_help = ""
        if self.comp_config is not None:
            _tab = self.comp_config.get("tab") or ".tab"
            _enter = self.comp_config.get("enter") or ".enter"
            _exit = self.comp_config.get("exit") or ".exit"
            disables = self.comp_config.get("disables", set())
            hides = self.comp_config.get("hides", set())
            hide_tabs = self.comp_config.get("hide_tabs", False)
            if self.comp_config.get("lite", False):
                hide_tabs = True
                hides = {"tab", "enter", "exit"}
            hides |= disables
            if len(hides) < 3:
                template = f"\n\n{{}}{{}}{{}}{lang.require('comp/nonebot', 'other')}\n"
                self._comp_help = template.format(
                    ((lang.require("comp/nonebot", "tab").format(cmd=_tab) + "\n") if "tab" not in hides else ""),
                    ((lang.require("comp/nonebot", "enter").format(cmd=_enter) + "\n") if "enter" not in hides else ""),
                    ((lang.require("comp/nonebot", "exit").format(cmd=_exit) + "\n") if "exit" not in hides else ""),
                )

            async def _waiter_handle(_bot: Bot, _event: Event, _matcher: Matcher, content: Message = EventMessage()):
                msg = str(content).lstrip()
                _future = self._futures[_bot.self_id][_event.get_session_id()]
                _interface = self._interfaces[_event.get_session_id()]
                if msg.startswith(_exit) and "exit" not in disables:
                    if msg == _exit:
                        _future.set_result(False)
                        await _matcher.finish()
                    else:
                        _future.set_result(None)
                        await _matcher.pause(
                            lang.require("analyser", "param_unmatched").format(target=msg.replace(_exit, "", 1))
                        )
                elif msg.startswith(_enter) and "enter" not in disables:
                    if msg == _enter:
                        _future.set_result(True)
                        await _matcher.finish()
                    else:
                        _future.set_result(None)
                        await _matcher.pause(
                            lang.require("analyser", "param_unmatched").format(target=msg.replace(_enter, "", 1))
                        )
                elif msg.startswith(_tab) and "tab" not in disables:
                    offset = msg.replace(_tab, "", 1).lstrip() or 1
                    try:
                        offset = int(offset)
                    except ValueError:
                        _future.set_result(None)
                        await _matcher.pause(lang.require("analyser", "param_unmatched").format(target=offset))
                    else:
                        _interface.tab(offset)
                        await _matcher.pause(
                            f"* {_interface.current()}" if hide_tabs else "\n".join(_interface.lines())
                        )
                else:
                    _future.set_result(content)
                    await _matcher.finish()

            self._waiter = _waiter_handle

    def __repr__(self) -> str:
        return f"Alconna(command={self.command!r})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, AlconnaRule) and self.command.path == other.command.path

    def __hash__(self) -> int:
        return hash(self.command.__hash__())

    async def handle(self, bot: Bot, event: Event, state: T_State, msg: Message) -> Union[Arparma, Literal[False]]:
        ctx = await self.executor.context_provider(event, bot, state)
        if self.comp_config is None:
            return self.command.parse(msg, ctx)
        res = None
        session_id = event.get_session_id()
        if session_id not in self._interfaces:
            self._interfaces[session_id] = CompSession(self.command)
        with self._interfaces[session_id]:
            res = self.command.parse(msg, ctx)
        if res:
            self._interfaces[session_id].exit()
            del self._interfaces[session_id]
            return res
        if not await self.executor.permission_check(bot, event):
            return False

        def _checker(_event: Event):
            return session_id == _event.get_session_id()

        self._matchers[session_id] = on_message(priority=0, block=True, rule=Rule(_checker), handlers=[self._waiter])
        res = Arparma(
            self.command.path,
            msg,
            False,
            error_info=SpecialOptionTriggered("completion"),
        )
        _futures = self._futures.setdefault(bot.self_id, {})
        _futures[session_id] = asyncio.get_running_loop().create_future()
        while self._interfaces[session_id].available:
            await self.send(f"{str(self._interfaces[session_id])}{self._comp_help}", bot, event, res)
            while True:
                try:
                    await asyncio.wait_for(_futures[session_id], timeout=self.comp_config.get("timeout", 60))
                except asyncio.TimeoutError:
                    await self.send(lang.require("comp/nonebot", "timeout"), bot, event, res)
                    self._interfaces[session_id].exit()
                    self._matchers[session_id].destroy()
                    del _futures[session_id]
                    del self._matchers[session_id]
                    del self._interfaces[session_id]
                    return res
                finally:
                    if not _futures[session_id].done():
                        _futures[session_id].cancel()
                ans: Union[Message, bool, None] = _futures[session_id].result()
                _futures[session_id] = asyncio.get_running_loop().create_future()
                if ans is False:
                    await self.send(lang.require("comp/nonebot", "exited"), bot, event, res)
                    self._interfaces[session_id].exit()
                    self._matchers[session_id].destroy()
                    del _futures[session_id]
                    del self._matchers[session_id]
                    del self._interfaces[session_id]
                    return res
                elif ans is None:
                    continue
                _res = self._interfaces[session_id].enter(None if ans is True else ans)
                if _res.result:
                    res = _res.result
                elif _res.exception and not isinstance(_res.exception, SpecialOptionTriggered):
                    await self.send(str(_res.exception), bot, event, res)
                break
        self._interfaces[session_id].exit()
        self._matchers[session_id].destroy()
        del _futures[session_id]
        del self._matchers[session_id]
        del self._interfaces[session_id]
        return res

    async def __call__(self, event: Event, state: T_State, bot: Bot) -> bool:
        self.executor.select(bot, event)
        if not (msg := await self.executor.message_provider(event, state, bot, self.use_origin)):
            return False
        msg = await self.executor.receive_wrapper(bot, event, msg)
        if isinstance(msg, UniMessage):
            msg = await msg.export(bot, fallback=True)
        Arparma._additional.update(bot=lambda: bot, event=lambda: event, state=lambda: state)
        adapter_name = bot.adapter.get_name()
        if adapter_name in MAPPING and MAPPING[adapter_name] not in _modules:
            importlib.import_module(f"nonebot_plugin_alconna.adapters.{MAPPING[adapter_name]}")
        with output_manager.capture(self.command.name) as cap:
            output_manager.set_action(lambda x: x, self.command.name)
            try:
                arp = await self.handle(bot, event, state, msg)
                if arp is False:
                    return False
            except Exception as e:
                arp = Arparma(self.command.path, msg, False, error_info=e)
            may_help_text: Optional[str] = cap.get("output", None)
        if not arp.head_matched:
            return False
        if not arp.matched and not may_help_text and self.skip:
            log(
                "TRACE",
                escape_tag(lang.require("nbp-alc", "log.parse").format(msg=msg, cmd=self.command.path, arp=arp)),
            )
            return False
        if arp.head_matched:
            log(
                "DEBUG",
                escape_tag(lang.require("nbp-alc", "log.parse").format(msg=msg, cmd=self.command.path, arp=arp)),
            )
        if not may_help_text and arp.error_info:
            may_help_text = repr(arp.error_info)
        if self.auto_send and may_help_text:
            await self.send(may_help_text, bot, event, arp)
            return False
        if not await self.executor.permission_check(bot, event):
            return False
        await self.executor.parse_wrapper(bot, state, event, arp)
        state[ALCONNA_RESULT] = CommandResult(self.command, arp, may_help_text)
        state[ALCONNA_EXEC_RESULT] = self.command.exec_result
        state[ALCONNA_EXTENSION] = self.executor.context
        return True

    async def send(self, text: str, bot: Bot, event: Event, arp: Arparma) -> Message:
        _t = str(arp.error_info) if isinstance(arp.error_info, SpecialOptionTriggered) else "help"
        try:
            msg = await self.executor.output_converter(_t, text)  # type: ignore
            if not msg:
                return await bot.send(event, text)
            if isinstance(msg, UniMessage):
                msg = await msg.export(bot, fallback=True)
            return await bot.send(event, msg)  # type: ignore
        except NotImplementedError:
            return await bot.send(event, event.get_message().__class__(text))


@init_spec(AlconnaRule)
def alconna(rule: AlconnaRule) -> Rule:
    return Rule(rule)



from nonebot.adapters.console.message import MessageSegment

from nonebot_plugin_alconna.typings import SegmentPattern

Emoji = SegmentPattern("emoji", MessageSegment, MessageSegment.emoji)
Markup = SegmentPattern("markup", MessageSegment, MessageSegment.markup)
Markdown = SegmentPattern("markdown", MessageSegment, MessageSegment.markdown)

Text = str



from __future__ import annotations

import asyncio
import weakref
from weakref import ref
from types import FunctionType
from typing_extensions import Self
from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    Generic,
    Literal,
    TypeVar,
    Callable,
    ClassVar,
    Hashable,
    Iterable,
    NoReturn,
    Protocol,
    Awaitable,
    cast,
    overload,
)

from nonebot.rule import Rule
from nonebot.params import Depends
from nonebot.utils import escape_tag
from nonebot.permission import Permission
from nonebot.dependencies import Dependent
from nepattern import ANY, STRING, AnyString
from nonebot.message import run_postprocessor
from nonebot.consts import ARG_KEY, RECEIVE_KEY
from nonebot.internal.params import DefaultParam
from arclet.alconna.typing import ShortcutRegWrapper
from tarina import lang, is_awaitable, run_always_await
from _weakref import _remove_dead_weakref  # type: ignore
from arclet.alconna.tools import AlconnaFormat, AlconnaString
from arclet.alconna.tools.construct import FuncMounter, MountConfig
from nonebot.plugin.on import on_message, store_matcher, get_matcher_source
from arclet.alconna import Arg, Args, Alconna, ShortcutArgs, command_manager
from nonebot.exception import PausedException, FinishedException, RejectedException
from nonebot.internal.adapter import Bot, Event, Message, MessageSegment, MessageTemplate
from nonebot.matcher import Matcher, matchers, current_bot, current_event, current_matcher
from nonebot.typing import T_State, T_Handler, T_RuleChecker, T_PermissionChecker, _DependentCallable

from .rule import alconna
from .typings import MReturn
from .model import CompConfig
from .uniseg import Segment, UniMessage
from .uniseg.template import UniMessageTemplate
from .extension import Extension, ExtensionExecutor
from .consts import ALCONNA_RESULT, ALCONNA_ARG_KEY, log
from .params import (
    CHECK,
    MIDDLEWARE,
    Check,
    AlconnaParam,
    AlcExecResult,
    ExtensionParam,
    assign,
    _seminal,
    _Dispatch,
    merge_path,
)

_M = Union[str, Message, MessageSegment, MessageTemplate, Segment, UniMessage, UniMessageTemplate]


class ArgsMounter(Protocol):
    args: Args


def extract_arg(path: str, target: ArgsMounter) -> Arg | None:
    """从 Alconna 中提取参数"""
    parts = path.split(".")
    if len(parts) == 1:
        return next((arg for arg in target.args.argument if arg.name == path), None)
    _parts, end = parts[:-1], parts[-1]
    if not (options := getattr(target, "options", None)):
        return
    for opt in options:
        if opt.dest == _parts[0]:
            return extract_arg(".".join(_parts[1:] + [end]), opt)
    return


def _validate(target: Arg[Any], arg: MessageSegment):
    value = target.value
    if value == ANY:
        return arg
    if value == AnyString or (value == STRING and arg.is_text()):
        return str(arg)
    default_val = target.field.default
    if arg.is_text():
        arg = arg.data["text"]
    res = value.validate(arg, default_val)
    if target.optional and res.flag != "valid":
        return
    if res.flag == "error":
        return
    return res._value  # noqa


class _method:
    def __init__(self, func: FunctionType):
        self.__func__ = func

    def __get__(self, instance, owner):
        if instance is None:
            return self.__func__.__get__(owner, owner)
        return self.__func__.__get__(instance, owner)


R = TypeVar("R")
R1 = TypeVar("R1")
T = TypeVar("T")
T1 = TypeVar("T1")


class WaiterIterator(Generic[R, T]):
    def __init__(self, waiter: Waiter[R], default: T, timeout: float = 120):
        self.waiter = waiter
        self.timeout = timeout
        self.default = default

    def __aiter__(self) -> Self:
        return self

    @overload
    def __anext__(self: WaiterIterator[R1, None]) -> Awaitable[R1 | None]: ...

    @overload
    def __anext__(self: WaiterIterator[R1, T1]) -> Awaitable[R1 | T1]: ...

    def __anext__(self):  # type: ignore
        return self.waiter.wait(default=self.default, timeout=self.timeout)  # type: ignore


class Waiter(Generic[R]):
    future: asyncio.Future
    handler: _DependentCallable[R]

    def __init__(self, handler: _DependentCallable[R], params: tuple, parameterless: Iterable[Any] | None = None):
        self.future = asyncio.Future()
        _handler = Dependent[Any].parse(call=handler, parameterless=parameterless, allow_types=params)

        async def wrapper(matcher: Matcher, bot: Bot, event: Event, state: T_State):
            if self.future.done():
                matcher.skip()
            result = await _handler(
                matcher=self,
                bot=bot,
                event=event,
                state=state,
            )
            if result is not None and not self.future.done():
                self.future.set_result(result)
                matcher.stop_propagation()
                await matcher.finish()
            matcher.skip()

        self.handler = wrapper

    def __aiter__(self) -> WaiterIterator[R, None]:
        return WaiterIterator(self, None)

    @overload
    def __call__(self, *, default: T, timeout: float = 120) -> WaiterIterator[R, T]: ...

    @overload
    def __call__(self, *, timeout: float = 120) -> WaiterIterator[R, None]: ...

    def __call__(
        self, *, default: T | None = None, timeout: float = 120
    ) -> WaiterIterator[R, T] | WaiterIterator[R, None]:
        return WaiterIterator(self, default, timeout)  # type: ignore

    @overload
    async def wait(self, *, default: R | T, timeout: float = 120) -> R | T: ...

    @overload
    async def wait(self, *, timeout: float = 120) -> R | None: ...

    async def wait(self, *, default: R | T | None = None, timeout: float = 120) -> R | T | None:
        matcher = on_message(priority=0, block=False, handlers=[self.handler])
        try:
            return await asyncio.wait_for(self.future, timeout)
        except asyncio.TimeoutError:
            return default
        finally:
            self.future = asyncio.Future()
            try:
                matcher.destroy()
            except (IndexError, ValueError):
                pass


class AlconnaMatcher(Matcher):
    command: ClassVar[Alconna]
    basepath: ClassVar[str]
    executor: ClassVar[ExtensionExecutor]

    @classmethod
    @overload
    def shortcut(cls, key: str, args: ShortcutArgs | None = None) -> str:
        """操作快捷命令

        Args:
            key (str): 快捷命令名
            args (ShortcutArgs[TDC]): 快捷命令参数, 不传入时则尝试使用最近一次使用的命令

        Returns:
            str: 操作结果

        Raises:
            ValueError: 快捷命令操作失败时抛出
        """
        ...

    @classmethod
    @overload
    def shortcut(
        cls,
        key: str,
        *,
        command: str | None = None,
        arguments: list[Any] | None = None,
        fuzzy: bool = True,
        prefix: bool = False,
        wrapper: ShortcutRegWrapper | None = None,
    ) -> str:
        """操作快捷命令

        Args:
            key (str): 快捷命令名
            command (TDC): 快捷命令指向的命令
            arguments (list[Any] | None, optional): 快捷命令参数, 默认为 `None`
            fuzzy (bool, optional): 是否允许命令后随参数, 默认为 `True`
            prefix (bool, optional): 是否调用时保留指令前缀, 默认为 `False`
            wrapper (ShortcutRegWrapper, optional): 快捷指令的正则匹配结果的额外处理函数, 默认为 `None`

        Returns:
            str: 操作结果

        Raises:
            ValueError: 快捷命令操作失败时抛出
        """
        ...

    @classmethod
    @overload
    def shortcut(cls, key: str, *, delete: Literal[True]) -> str:
        """操作快捷命令

        Args:
            key (str): 快捷命令名
            delete (bool): 是否删除快捷命令

        Returns:
            str: 操作结果

        Raises:
            ValueError: 快捷命令操作失败时抛出
        """
        ...

    @classmethod
    def shortcut(cls, key: str, args: ShortcutArgs | None = None, delete: bool = False, **kwargs):
        """操作快捷命令

        Args:
            key (str): 快捷命令名
            args (ShortcutArgs[TDC] | None, optional): 快捷命令参数, 不传入时则尝试使用最近一次使用的命令
            delete (bool, optional): 是否删除快捷命令, 默认为 `False`
            command (TDC, optional): 快捷命令指向的命令
            arguments (list[Any] | None, optional): 快捷命令参数, 默认为 `None`
            fuzzy (bool, optional): 是否允许命令后随参数, 默认为 `True`
            prefix (bool, optional): 是否调用时保留指令前缀, 默认为 `False`
            wrapper (ShortcutRegWrapper, optional): 快捷指令的正则匹配结果的额外处理函数, 默认为 `None`

        Returns:
            str: 操作结果

        Raises:
            ValueError: 快捷命令操作失败时抛出
        """
        return cls.command.shortcut(key, args, delete, **kwargs)  # type: ignore

    if TYPE_CHECKING:

        @classmethod
        def set_path_arg(cls_or_self, path: str, content: Any) -> None:  # type: ignore
            ...

        @classmethod
        def get_path_arg(cls_or_self, path: str, default: Any) -> Any:  # type: ignore
            ...

    else:

        @_method
        def set_path_arg(cls_or_self, path: str, content: Any) -> None:
            """设置一个 `got_path` 内容"""
            if isinstance(cls_or_self, AlconnaMatcher):
                state = cls_or_self.state
            else:
                state = current_matcher.get().state
            state[ALCONNA_ARG_KEY.format(key=merge_path(path, cls_or_self.basepath))] = content

        @_method
        def get_path_arg(cls_or_self, path: str, default: Any) -> Any:
            """获取一个 `got_path` 内容"""
            if isinstance(cls_or_self, AlconnaMatcher):
                state = cls_or_self.state
            else:
                state = current_matcher.get().state
            return state.get(ALCONNA_ARG_KEY.format(key=merge_path(path, cls_or_self.basepath)), default)

    @classmethod
    def assign(
        cls,
        path: str,
        value: Any = _seminal,
        or_not: bool = False,
        additional: CHECK | None = None,
        parameterless: Iterable[Any] | None = None,
    ) -> Callable[[T_Handler], T_Handler]:
        """装饰一个函数来向事件响应器直接添加一个处理函数

        此装饰是 @handle([Check(assign(...))]) 的快捷写法

        参数:
            path: 指定的查询路径;
                "$main" 表示没有任何选项/子命令匹配的时候;
                "~XX" 时会把 "~" 替换为父级 assign 的 path
            value: 可能的指定查询值
            or_not: 是否同时处理没有查询成功的情况
            additional: 额外的检查函数
            parameterless: 非参数类型依赖列表
        """
        parameterless = [
            Check(assign(merge_path(path, cls.basepath), value, or_not, additional)),
            *(parameterless or []),
        ]

        def _decorator(func: T_Handler) -> T_Handler:
            cls.append_handler(func, parameterless=parameterless)
            return func

        return _decorator

    @classmethod
    def dispatch(
        cls,
        path: str,
        value: Any = _seminal,
        or_not: bool = False,
        additional: CHECK | None = None,
        rule: Rule | T_RuleChecker | None = None,
        permission: Permission | T_PermissionChecker | None = None,
        *,
        handlers: list[T_Handler | Dependent] | None = None,
        temp: bool = False,
        expire_time: datetime | timedelta | None = None,
        priority: int = 1,
        block: bool = False,
        state: T_State | None = None,
        _depth: int = 0,
    ) -> type[AlconnaMatcher]:
        """注册一个消息事件响应器，
        并且当消息由指定 Alconna 解析并且结果符合 assign 预期时执行

        参数:
            path: 指定的查询路径;
                "$main" 表示没有任何选项/子命令匹配的时候;
                "~XX" 时会把 "~" 替换为父级 assign 的 path
            value: 可能的指定查询值
            or_not: 是否同时处理没有查询成功的情况
            additional: 额外的检查函数
            rule: 事件响应规则
            permission: 事件响应权限
            handlers: 事件处理函数列表
            temp: 是否为临时事件响应器（仅执行一次）
            expire_time: 事件响应器最终有效时间点，过时即被删除
            priority: 事件响应器优先级，其会基于父级事件响应器的优先级进行累加
            block: 是否阻止事件向更低优先级传递
            state: 默认 state
        """

        fn = _Dispatch(merge_path(path, cls.basepath), value, or_not, additional)
        cls.handle()(fn.set)

        matcher: type[AlconnaMatcher] = cls.new(
            "",
            rule & Rule(fn),
            Permission() | permission,
            temp=temp,
            expire_time=expire_time,
            priority=cls.priority + priority,
            block=block,
            handlers=handlers,
            source=get_matcher_source(_depth + 1),
            default_state=state,
        )
        store_matcher(matcher)
        matcher.command = cls.command
        matcher.basepath = merge_path(path, cls.basepath)
        matcher.executor = cls.executor
        return matcher

    @classmethod
    def handle(
        cls,
        parameterless: Iterable[Any] | None = None,
        override: tuple[Literal["insert", "replace"], int] | None = None,
    ) -> Callable[[T_Handler], T_Handler]:
        """装饰一个函数来向事件响应器直接添加一个处理函数

        参数:
            parameterless: 非参数类型依赖列表
            override: 是否定制优先级
        """

        def _decorator(func: T_Handler) -> T_Handler:
            if override is None:
                cls.append_handler(func, parameterless=parameterless)
            else:
                handler_ = Dependent[Any].parse(
                    call=func,
                    parameterless=parameterless,
                    allow_types=cls.HANDLER_PARAM_TYPES,
                )
                if override[0] == "insert":
                    cls.handlers.insert(override[1], handler_)
                else:
                    cls.handlers[override[1]] = handler_
            return func

        return _decorator

    @classmethod
    def receive(
        cls,
        id: str = "",
        parameterless: Iterable[Any] | None = None,
        override: tuple[Literal["insert", "replace"], int] | None = None,
    ) -> Callable[[T_Handler], T_Handler]:
        """装饰一个函数来指示 NoneBot 在接收用户新的一条消息后继续运行该函数

        参数:
            id: 消息 ID，若不指定则使用当前消息 ID
            parameterless: 非参数类型依赖列表
            override: 是否定制优先级
        """

        async def _receive(event: Event, matcher: Matcher) -> None | NoReturn:
            nonlocal id
            if not id:
                try:
                    id = UniMessage.get_message_id(event, current_bot.get())
                except Exception:
                    pass
            matcher.set_target(RECEIVE_KEY.format(id=id))
            if matcher.get_target() == RECEIVE_KEY.format(id=id):
                matcher.set_receive(id, event)
                return
            if matcher.get_receive(id, ...) is not ...:
                return
            await matcher.reject()

        _parameterless = (Depends(_receive), *(parameterless or ()))

        def _decorator(func: T_Handler) -> T_Handler:
            if cls.handlers and cls.handlers[-1].call is func:
                func_handler = cls.handlers[-1]
                new_handler = Dependent(
                    call=func_handler.call,
                    params=func_handler.params,
                    parameterless=Dependent.parse_parameterless(tuple(_parameterless), cls.HANDLER_PARAM_TYPES)
                    + func_handler.parameterless,
                )
                cls.handlers[-1] = new_handler
            elif override is None:
                cls.append_handler(func, parameterless=_parameterless)
            else:
                handler_ = Dependent[Any].parse(
                    call=func,
                    parameterless=_parameterless,
                    allow_types=cls.HANDLER_PARAM_TYPES,
                )
                mode, index = override
                if mode == "insert":
                    cls.handlers.insert(index, handler_)
                else:
                    cls.handlers[index] = handler_

            return func

        return _decorator

    @classmethod
    def got(
        cls,
        key: str,
        prompt: _M | None = None,
        parameterless: Iterable[Any] | None = None,
        override: tuple[Literal["insert", "replace"], int] | None = None,
    ) -> Callable[[T_Handler], T_Handler]:
        """装饰一个函数来指示 NoneBot 获取一个参数 `key`

        当要获取的 `key` 不存在时接收用户新的一条消息再运行该函数，
        如果 `key` 已存在则直接继续运行

        参数:
            key: 参数名
            prompt: 在参数不存在时向用户发送的消息, 支持 `UniMessage`
            parameterless: 非参数类型依赖列表
            override: 是否定制优先级
        """

        async def _key_getter(event: Event, matcher: AlconnaMatcher, bot: Bot, state: T_State):
            matcher.set_target(ARG_KEY.format(key=key))
            if matcher.get_target() == ARG_KEY.format(key=key):
                msg = await cls.executor.select(bot, event).message_provider(event, state, bot)
                if not msg:
                    await matcher.reject(prompt, fallback=True)
                if isinstance(msg, UniMessage):
                    msg = await msg.export(bot, True)
                matcher.set_arg(key, msg)
                return
            if matcher.get_arg(key, ...) is not ...:
                return
            await matcher.reject(prompt, fallback=True)

        _parameterless = (Depends(_key_getter), *(parameterless or ()))

        def _decorator(func: T_Handler) -> T_Handler:
            if cls.handlers and cls.handlers[-1].call is func:
                func_handler = cls.handlers[-1]
                new_handler = Dependent(
                    call=func_handler.call,
                    params=func_handler.params,
                    parameterless=Dependent.parse_parameterless(tuple(_parameterless), cls.HANDLER_PARAM_TYPES)
                    + func_handler.parameterless,
                )
                cls.handlers[-1] = new_handler
            elif override is None:
                cls.append_handler(func, parameterless=_parameterless)
            else:
                handler_ = Dependent[Any].parse(
                    call=func,
                    parameterless=_parameterless,
                    allow_types=cls.HANDLER_PARAM_TYPES,
                )
                mode, index = override
                if mode == "insert":
                    cls.handlers.insert(index, handler_)
                else:
                    cls.handlers[index] = handler_
            return func

        return _decorator

    @classmethod
    def got_path(
        cls,
        path: str,
        prompt: _M | None = None,
        middleware: MIDDLEWARE | None = None,
        parameterless: Iterable[Any] | None = None,
        override: tuple[Literal["insert", "replace"], int] | None = None,
    ) -> Callable[[T_Handler], T_Handler]:
        """装饰一个函数来指示 NoneBot 获取一个路径下的参数 `path`

        当要获取的 `path` 不存在时接收用户新的一条消息再运行该函数

        如果 `path` 已存在则直接继续运行

        本插件会获取消息的最后一个消息段并转为 path 对应的类型

        参数:
            path: 参数路径名; "~XX" 时会把 "~" 替换为父级 assign 的 path
            prompt: 在参数不存在时向用户发送的消息，支持 `UniMessage`
            parameterless: 非参数类型依赖列表
            override: 是否定制优先级
        """
        path = merge_path(path, cls.basepath)
        if not (arg := extract_arg(path, cls.command)):
            raise ValueError(lang.require("nbp-alc", "error.matcher_got_path").format(path=path, cmd=cls.command.path))

        async def _key_getter(event: Event, bot: Bot, matcher: AlconnaMatcher, state: T_State):
            matcher.set_target(ALCONNA_ARG_KEY.format(key=path))
            if matcher.get_target() == ALCONNA_ARG_KEY.format(key=path):
                msg = await cls.executor.select(bot, event).message_provider(event, state, bot)
                if not msg:
                    await matcher.reject(prompt, fallback=True)
                if isinstance(msg, UniMessage):
                    msg = await msg.export(bot, True)
                ms = msg[0]
                if ms.is_text() and not ms.data["text"].strip():
                    await matcher.reject(prompt, fallback=True)
                log("DEBUG", escape_tag(lang.require("nbp-alc", "log.got_path/ms").format(path=path, ms=ms)))
                if (res := _validate(arg, ms)) is None:  # type: ignore
                    log(
                        "TRACE",
                        escape_tag(lang.require("nbp-alc", "log.got_path/validate").format(path=path, validate=res)),
                    )
                    await matcher.reject(prompt, fallback=True)
                if middleware:
                    res = await run_always_await(middleware, event, bot, matcher.state, res)
                matcher.set_path_arg(path, res)
                return
            if matcher.state.get(ALCONNA_ARG_KEY.format(key=path), ...) is not ...:
                return
            await matcher.reject(prompt, fallback=True)

        _parameterless = (*(parameterless or ()), Depends(_key_getter))

        def _decorator(func: T_Handler) -> T_Handler:
            if cls.handlers and cls.handlers[-1].call is func:
                func_handler = cls.handlers[-1]
                new_handler = Dependent(
                    call=func_handler.call,
                    params=func_handler.params,
                    parameterless=Dependent.parse_parameterless(tuple(_parameterless), cls.HANDLER_PARAM_TYPES)
                    + func_handler.parameterless,
                )
                cls.handlers[-1] = new_handler
            elif override is None:
                cls.append_handler(func, parameterless=_parameterless)
            else:
                handler_ = Dependent[Any].parse(
                    call=func,
                    parameterless=_parameterless,
                    allow_types=cls.HANDLER_PARAM_TYPES,
                )
                mode, index = override
                if mode == "insert":
                    cls.handlers.insert(index, handler_)
                else:
                    cls.handlers[index] = handler_

            return func

        return _decorator

    @classmethod
    def convert(cls, message: _M) -> Message | UniMessage | str:
        bot = current_bot.get()
        event = current_event.get()
        state = current_matcher.get().state
        if isinstance(message, MessageTemplate):
            return message.format(**state[ALCONNA_RESULT].result.all_matched_args, **state)
        if isinstance(message, UniMessageTemplate):
            extra = {"$event": event, "$bot": bot, "$target": UniMessage.get_target(event, bot)}
            try:
                msg_id = UniMessage.get_message_id(event, bot)
            except Exception:
                pass
            else:
                extra["$message_id"] = msg_id
            return message.format(**state[ALCONNA_RESULT].result.all_matched_args, **state, **extra)
        if isinstance(message, Segment):
            return UniMessage(message)
        if isinstance(message, MessageSegment):
            return message.get_message_class()(message)
        return message

    @classmethod
    async def send(
        cls,
        message: _M,
        fallback: bool = False,
        **kwargs: Any,
    ) -> Any:
        """发送一条消息给当前交互用户

        `AlconnaMatcher` 下增加了对 UniMessage 的发送支持

        参数:
            message: 消息内容, 支持 `UniMessage`
            fallback: 若 UniMessage 中的元素无法被解析为当前 adapter 的消息元素时，
                是否转为字符串
            kwargs: {ref}`nonebot.adapters.Bot.send` 的参数，
                请参考对应 adapter 的 bot 对象 api
        """
        bot = current_bot.get()
        event = current_event.get()
        _message = await cls.executor.send_wrapper(bot, event, cls.convert(message))
        if isinstance(_message, UniMessage):
            res = await _message.export(bot, fallback)
        else:
            res = _message
        return await bot.send(event=event, message=res, **kwargs)

    @classmethod
    async def finish(
        cls,
        message: _M | None = None,
        fallback: bool = False,
        **kwargs,
    ) -> NoReturn:
        """发送一条消息给当前交互用户并结束当前事件响应器

        参数:
            message: 消息内容, 支持 `UniMessage`
            fallback: 若 UniMessage 中的元素无法被解析为当前 adapter 的消息元素时，
                是否转为字符串
            kwargs: {ref}`nonebot.adapters.Bot.send` 的参数，
                请参考对应 adapter 的 bot 对象 api
        """
        if message is not None:
            await cls.send(message, fallback=fallback, **kwargs)
        raise FinishedException

    @classmethod
    async def pause(
        cls,
        prompt: _M | None = None,
        fallback: bool = False,
        **kwargs,
    ) -> NoReturn:
        """发送一条消息给当前交互用户并暂停事件响应器，在接收用户新的一条消息后继续下一个处理函数

        参数:
            prompt: 消息内容, 支持 `UniMessage`
            fallback: 若 UniMessage 中的元素无法被解析为当前 adapter 的消息元素时，
                是否转为字符串
            kwargs: {ref}`nonebot.adapters.Bot.send` 的参数，
                请参考对应 adapter 的 bot 对象 api
        """
        if prompt is not None:
            await cls.send(prompt, fallback=fallback, **kwargs)
        raise PausedException

    @classmethod
    async def reject(
        cls,
        prompt: _M | None = None,
        fallback: bool = False,
        **kwargs,
    ) -> NoReturn:
        """最近使用 `got` / `receive` 接收的消息不符合预期，
        发送一条消息给当前交互用户并将当前事件处理流程中断在当前位置，在接收用户新的一个事件后从头开始执行当前处理函数

        参数:
            prompt: 消息内容, 支持 `UniMessage`
            fallback: 若 UniMessage 中的元素无法被解析为当前 adapter 的消息元素时，
                是否转为字符串
            kwargs: {ref}`nonebot.adapters.Bot.send` 的参数，
                请参考对应 adapter 的 bot 对象 api
        """
        if prompt is not None:
            await cls.send(prompt, fallback=fallback, **kwargs)
        raise RejectedException

    @classmethod
    async def reject_path(
        cls,
        path: str,
        prompt: _M | None = None,
        fallback: bool = False,
        **kwargs,
    ) -> NoReturn:
        """最近使用 `got_path` 接收的消息不符合预期，
        发送一条消息给当前交互用户并将当前事件处理流程中断在当前位置，在接收用户新的一条消息后从头开始执行当前处理函数

        参数:
            path: 参数路径名; "~XX" 时会把 "~" 替换为父级 assign 的 path
            prompt: 消息内容, 支持 `UniMessage`
            fallback: 若 UniMessage 中的元素无法被解析为当前 adapter 的消息元素时，
                是否转为字符串
            kwargs: {ref}`nonebot.adapters.Bot.send` 的参数，
                请参考对应 adapter 的 bot 对象 api
        """
        matcher = current_matcher.get()
        matcher.set_target(ALCONNA_ARG_KEY.format(key=merge_path(path, cls.basepath)))
        if prompt is not None:
            await cls.send(prompt, fallback=fallback, **kwargs)
        raise RejectedException

    @classmethod
    async def reject_arg(
        cls,
        key: str,
        prompt: _M | None = None,
        fallback: bool = False,
        **kwargs,
    ) -> NoReturn:
        """最近使用 `got` 接收的消息不符合预期，
        发送一条消息给当前交互用户并将当前事件处理流程中断在当前位置，在接收用户新的一条消息后从头开始执行当前处理函数

        参数:
            key: 参数名
            prompt: 消息内容, 支持 `UniMessage`
            fallback: 若 UniMessage 中的元素无法被解析为当前 adapter 的消息元素时，
                是否转为字符串
            kwargs: {ref}`nonebot.adapters.Bot.send` 的参数，
                请参考对应 adapter 的 bot 对象 api
        """
        matcher = current_matcher.get()
        matcher.set_target(ARG_KEY.format(key=key))
        if prompt is not None:
            await cls.send(prompt, fallback=fallback, **kwargs)
        raise RejectedException

    @classmethod
    async def reject_receive(
        cls,
        id: str = "",
        prompt: _M | None = None,
        fallback: bool = False,
        **kwargs,
    ) -> NoReturn:
        """最近使用 `receive` 接收的消息不符合预期，
        发送一条消息给当前交互用户并将当前事件处理流程中断在当前位置，在接收用户新的一个事件后从头开始执行当前处理函数

        参数:
            id: 消息 id
            prompt: 消息内容, 支持 `UniMessage`
            fallback: 若 UniMessage 中的元素无法被解析为当前 adapter 的消息元素时，
                是否转为字符串
            kwargs: {ref}`nonebot.adapters.Bot.send` 的参数，
                请参考对应 adapter 的 bot 对象 api
        """
        matcher = current_matcher.get()
        matcher.set_target(RECEIVE_KEY.format(id=id))
        if prompt is not None:
            await cls.send(prompt, fallback=fallback, **kwargs)
        raise RejectedException

    @classmethod
    def waiter(cls, parameterless: Iterable[Any] | None = None):
        """装饰一个函数来创建一个 `Waiter` 对象用以等待用户输入

        函数内需要自行判断输入是否符合预期并返回结果

        参数:
            parameterless: 非参数类型依赖列表
        """

        def wrapper(func: _DependentCallable[R]):
            return Waiter(func, cls.HANDLER_PARAM_TYPES, parameterless)

        return wrapper


def on_alconna(
    command: Alconna | str,
    rule: Rule | T_RuleChecker | None = None,
    skip_for_unmatch: bool = True,
    auto_send_output: bool = False,
    aliases: set[str] | tuple[str, ...] | None = None,
    comp_config: CompConfig | None = None,
    extensions: list[type[Extension] | Extension] | None = None,
    exclude_ext: list[type[Extension] | str] | None = None,
    use_origin: bool = False,
    use_cmd_start: bool = False,
    use_cmd_sep: bool = False,
    permission: Permission | T_PermissionChecker | None = None,
    *,
    handlers: list[T_Handler | Dependent] | None = None,
    temp: bool = False,
    expire_time: datetime | timedelta | None = None,
    priority: int = 1,
    block: bool = False,
    default_state: T_State | None = None,
    _depth: int = 0,
) -> type[AlconnaMatcher]:
    """注册一个事件响应器，并且当消息由指定 Alconna 解析并传出有效结果时响应。

    参数:
        command: Alconna 命令
        rule: 事件响应规则
        skip_for_unmatch: 是否在解析失败时跳过
        auto_send_output: 是否自动发送输出信息并跳过
        aliases: 命令别名
        comp_config: 补全会话配置, 不传入则不启用补全会话
        extensions: 需要加载的匹配扩展
        exclude_ext: 需要排除的匹配扩展
        use_origin: 是否使用未经 to_me 等处理过的消息
        use_cmd_start: 是否使用 nb 全局配置里的命令起始符
        use_cmd_sep: 是否使用 nb 全局配置里的命令分隔符
        permission: 事件响应权限
        handlers: 事件处理函数列表
        temp: 是否为临时事件响应器（仅执行一次）
        expire_time: 事件响应器最终有效时间点，过时即被删除
        priority: 事件响应器优先级
        block: 是否阻止事件向更低优先级传递
        state: 默认 state
    """
    if isinstance(command, str):
        command = AlconnaFormat(command)
    _rule = alconna(
        command,
        skip_for_unmatch,
        auto_send_output,
        comp_config,
        extensions,
        exclude_ext,
        use_origin,
        use_cmd_start,
        use_cmd_sep,
        aliases,
    )
    executor = cast(ExtensionExecutor, list(_rule.checkers)[0].call.executor)  # type: ignore
    params = (
        (ExtensionParam.new(executor),)
        + Matcher.HANDLER_PARAM_TYPES[:-1]
        + (
            AlconnaParam,
            DefaultParam,
        )
    )
    source = get_matcher_source(_depth)
    NewMatcher = type(
        AlconnaMatcher.__name__,
        (AlconnaMatcher,),
        {
            "_source": source,
            "type": "",
            "rule": _rule & rule,
            "permission": Permission() | permission,
            "handlers": (
                [
                    (
                        handler
                        if isinstance(handler, Dependent)
                        else Dependent[Any].parse(call=handler, allow_types=params)
                    )
                    for handler in handlers
                ]
                if handlers
                else []
            ),
            "temp": temp,
            "expire_time": (
                expire_time and (expire_time if isinstance(expire_time, datetime) else datetime.now() + expire_time)
            ),
            "priority": priority,
            "block": block,
            "_default_state": default_state or {},
            "_default_type_updater": None,
            "_default_permission_updater": None,
        },
    )
    matcher: type[AlconnaMatcher] = cast("type[AlconnaMatcher]", NewMatcher)
    matcher.HANDLER_PARAM_TYPES = params
    log("TRACE", f"Define new matcher {NewMatcher}")

    matchers[priority].append(NewMatcher)
    store_matcher(matcher)
    matcher.command = command
    matcher.basepath = ""
    matcher.executor = executor
    command.meta.extra["matcher.source"] = matcher._source
    command.meta.extra["matcher.position"] = (priority, len(matchers[priority]) - 1)

    def remove(wr, selfref=ref(command.meta), _atomic_removal=_remove_dead_weakref):
        self = selfref()
        if self is not None:
            _atomic_removal(self.extra, wr.key)

    command.meta.extra["matcher"] = weakref.KeyedRef(matcher, remove, "matcher")
    return matcher


def referent(cmd: str | Alconna) -> type[AlconnaMatcher]:
    """获取 Alconna 对应的 Matcher

    若不存在则返回一个新的 Matcher
    """
    if isinstance(cmd, str):
        try:
            cmd = command_manager.get_command(cmd)
        except ValueError:
            return on_alconna(cmd)
    if "matcher" in cmd.meta.extra:
        return cmd.meta.extra["matcher"]()
    return on_alconna(cmd)


def funcommand(
    name: str | None = None,
    prefixes: list[str] | None = None,
    description: str | None = None,
    skip_for_unmatch: bool = True,
    auto_send_output: bool = False,
    extensions: list[type[Extension] | Extension] | None = None,
    exclude_ext: list[type[Extension] | str] | None = None,
    use_origin: bool = False,
    use_cmd_start: bool = False,
    use_cmd_sep: bool = False,
    rule: Rule | T_RuleChecker | None = None,
    permission: Permission | T_PermissionChecker | None = None,
    *,
    handlers: list[T_Handler | Dependent] | None = None,
    temp: bool = False,
    expire_time: datetime | timedelta | None = None,
    priority: int = 1,
    block: bool = False,
    default_state: T_State | None = None,
    _depth: int = 0,
) -> Callable[[Callable[..., MReturn]], type[AlconnaMatcher]]:
    _config: MountConfig = {"raise_exception": False}
    if name:
        _config["command"] = name
    if prefixes:
        _config["prefixes"] = prefixes
    if description:
        _config["description"] = description

    def wrapper(func: Callable[..., MReturn]) -> type[AlconnaMatcher]:
        matcher = on_alconna(
            FuncMounter(func, _config),  # type: ignore
            rule,
            skip_for_unmatch,
            auto_send_output,
            extensions=extensions,
            exclude_ext=exclude_ext,
            use_origin=use_origin,
            use_cmd_start=use_cmd_start,
            use_cmd_sep=use_cmd_sep,
            permission=permission,
            handlers=handlers,
            temp=temp,
            expire_time=expire_time,
            priority=priority,
            block=block,
            default_state=default_state,
            _depth=_depth + 1,
        )

        @matcher.handle()
        async def handle_func(results: AlcExecResult):
            if res := results.get(func.__name__):
                if isinstance(res, Hashable) and is_awaitable(res):
                    res = await res
                if isinstance(res, (str, Message, MessageSegment, Segment, UniMessage, UniMessageTemplate)):
                    await matcher.send(res, fallback=True)

        return matcher

    return wrapper


class Command(AlconnaString):
    def build(
        self,
        rule: Rule | T_RuleChecker | None = None,
        skip_for_unmatch: bool = True,
        auto_send_output: bool = False,
        aliases: set[str] | tuple[str, ...] | None = None,
        comp_config: CompConfig | None = None,
        extensions: list[type[Extension] | Extension] | None = None,
        exclude_ext: list[type[Extension] | str] | None = None,
        use_origin: bool = False,
        use_cmd_start: bool = False,
        use_cmd_sep: bool = False,
        permission: Permission | T_PermissionChecker | None = None,
        *,
        handlers: list[T_Handler | Dependent] | None = None,
        temp: bool = False,
        expire_time: datetime | timedelta | None = None,
        priority: int = 1,
        block: bool = False,
        default_state: T_State | None = None,
        _depth: int = 0,
    ):
        params = locals().copy()
        params["_depth"] += 1
        params.pop("self")
        params.pop("__class__")
        alc = super().build()
        matcher = on_alconna(alc, **params)
        if self.actions:

            @matcher.handle()
            async def handle_actions(results: AlcExecResult):
                for res in results.values():
                    if isinstance(res, Hashable) and is_awaitable(res):
                        res = await res
                    if isinstance(res, (str, Message, MessageSegment, Segment, UniMessage, UniMessageTemplate)):
                        await matcher.send(res, fallback=True)

        return matcher


@run_postprocessor
def _exit_executor(matcher: AlconnaMatcher):
    matcher.executor.context.clear()



from __future__ import annotations

from typing_extensions import ParamSpec, TypeAlias
from typing import Any, Union, Generic, TypeVar, Callable, Awaitable

from tarina import lang
from arclet.alconna import Arparma
from nonebot.typing import T_State
from nepattern import MatchMode, BasePattern, MatchFailed
from nonebot.internal.adapter import Bot, Event, Message, MessageSegment

from .uniseg import Segment, UniMessage

T = TypeVar("T")
TMS = TypeVar("TMS", bound=MessageSegment)
TCallable = TypeVar("TCallable", bound=Callable[..., Any])
P = ParamSpec("P")


class SegmentPattern(BasePattern[TMS, MessageSegment], Generic[TMS, P]):
    def __init__(
        self,
        name: str,
        origin: type[TMS],
        call: Callable[P, TMS],
        additional: Callable[[TMS], bool] | None = None,
        handle: Callable[[TMS], TMS | None] | None = None,
    ):
        super().__init__(
            mode=MatchMode.TYPE_CONVERT,
            origin=origin,
            alias=name,
            accepts=MessageSegment,
            validators=[additional] if additional else [],
        )
        self.pattern = name
        self.call = call
        self.handle = handle

    def match(self, input_: MessageSegment) -> TMS:
        if not isinstance(input_, self.origin):
            raise MatchFailed(lang.require("nepattern", "type_error").format(target=type(input_)))
        if input_.type != self.pattern:
            raise MatchFailed(lang.require("nepattern", "content_error").format(target=input_))
        if self.handle:
            if res := self.handle(input_):
                return res
            raise MatchFailed(lang.require("nepattern", "content_error").format(target=input_))
        return input_

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> TMS:
        return self.call(*args, **kwargs)  # type: ignore


class TextSegmentPattern(BasePattern[TMS, Union[MessageSegment, str]], Generic[TMS, P]):
    def __init__(
        self,
        name: str,
        origin: type[TMS],
        call: Callable[P, TMS],
        locator: Callable[[str, str], bool] | None = None,
    ):
        super().__init__(
            mode=MatchMode.TYPE_CONVERT,
            origin=origin,
            alias=name,
            accepts=Union[MessageSegment, str],
        )
        self.pattern = name
        self.call = call
        self.locator = locator

    def match(self, input_: str | MessageSegment) -> TMS:
        if not isinstance(input_, (str, self.origin)):  # type: ignore
            raise MatchFailed(lang.require("nepattern", "type_error").format(target=type(input_)))
        if isinstance(input_, str):
            if self.locator and not self.locator(input_, self.pattern):
                raise MatchFailed(lang.require("nepattern", "content_error").format(target=input_))
            return self.call(input_)  # type: ignore
        if input_.type != self.pattern:
            raise MatchFailed(lang.require("nepattern", "content_error").format(target=input_))
        return input_

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> TMS:
        return self.call(*args, **kwargs)  # type: ignore


MReturn: TypeAlias = Union[
    Union[str, Segment, UniMessage, Message, MessageSegment],
    Awaitable[Union[str, Segment, UniMessage, Message, MessageSegment]],
]
MIDDLEWARE: TypeAlias = Callable[[Event, Bot, T_State, Any], Any]
CHECK: TypeAlias = Callable[[Event, Bot, T_State, Arparma], Awaitable[bool]]



from typing import List, Literal, Optional

from pydantic import Field, BaseModel


class Config(BaseModel):
    """Plugin Config Here"""

    alconna_auto_send_output: bool = False
    """是否全局启用输出信息自动发送"""

    alconna_use_command_start: bool = False
    """是否将 COMMAND_START 作为全局命令前缀"""

    alconna_auto_completion: bool = False
    """是否全局启用命令自动补全"""

    alconna_use_origin: bool = False
    """是否全局使用原始消息 (即未经过 to_me 等处理的)"""

    alconna_use_command_sep: bool = False
    """是否将 COMMAND_SEP 作为全局命令分隔符"""

    alconna_global_extensions: List[str] = Field(default_factory=list)
    """全局加载的扩展, 路径以 . 分隔, 如 foo.bar.baz:DemoExtension"""

    alconna_context_style: Optional[Literal["bracket", "parentheses"]] = Field(default=None)
    "全局命令上下文插值的风格，None 为关闭，bracket 为 {...}，parentheses 为 $(...)"



import pytest
from nonebug import App
from nonebot import get_adapter
from arclet.alconna import Alconna
from nonebot.adapters.onebot.v11 import Bot, Adapter, Message, MessageSegment

from tests.fake import fake_group_message_event_v11


def test_uniseg():
    from nonebot_plugin_alconna import Text, Other, Segment
    from nonebot_plugin_alconna.uniseg import FallbackSegment

    assert str(Other(FallbackSegment.text("123"))) == "[text]"
    assert str(Segment()) == "[segment]"
    assert str(Text("123")) == "123"


def test_unimsg():
    from nonebot_plugin_alconna.uniseg import FallbackSegment
    from nonebot_plugin_alconna import Text, Other, Segment, UniMessage

    msg = UniMessage([Other(FallbackSegment.text("123")), Segment(), Text("123")])
    assert str(msg) == "[text][segment]123"
    assert (
        repr(msg)
        == "[Other(origin=FallbackSegment(type='text', data={'text': '123'})), Segment(), Text(text='123', style=None)]"  # noqa: E501
    )
    assert (
        repr(UniMessage.text("123") + Other(FallbackSegment.text("123")))
        == "[Text(text='123', style=None), Other(origin=FallbackSegment(type='text', data={'text': '123'}))]"
    )


@pytest.mark.asyncio()
async def test_unimsg_template(app: App):
    from nonebot_plugin_alconna.uniseg import FallbackSegment
    from nonebot_plugin_alconna import At, Text, Other, UniMessage, on_alconna

    assert UniMessage.template("{} {}").format("hello", Other(FallbackSegment.text("123"))) == UniMessage(
        [Text("hello "), Other(FallbackSegment.text("123"))]
    )
    assert UniMessage.template("{:At(user, target)}").format(target="123") == UniMessage(At("user", "123"))
    assert UniMessage.template("{:At(flag=user, target=id)}").format(id="123") == UniMessage(At("user", "123"))
    assert UniMessage.template("{:At(flag=user, target=123)}").format() == UniMessage(At("user", "123"))

    matcher = on_alconna(Alconna("test_unimsg_template"))

    @matcher.handle()
    async def handle():
        await matcher.finish(UniMessage.template("{:Reply($message_id)}{:At(user, $event.get_user_id()[1:])}"))

    async with app.test_matcher(matcher) as ctx:
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter)
        event = fake_group_message_event_v11(message=Message("test_unimsg_template"), user_id=123)
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, MessageSegment.reply(1) + MessageSegment.at(23))
        ctx.should_finished(matcher)


@pytest.mark.asyncio()
async def test_unimsg_send(app: App):
    from nonebot_plugin_alconna import MsgId, Target, UniMessage, on_alconna

    matcher = on_alconna(Alconna("test_unimsg_send"))

    @matcher.handle()
    async def handle(msg: MsgId):
        receipt = await UniMessage("hello!").send(at_sender=True, reply_to=msg)
        receipt.msg_ids[0] = {"message_id": 2}
        await UniMessage("world!").send(at_sender=True, reply_to=receipt.get_reply())
        await receipt.recall(1)

    async with app.test_matcher(matcher) as ctx:
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter)
        event = fake_group_message_event_v11(message=Message("test_unimsg_send"), user_id=123, group_id=456)
        ctx.receive_event(bot, event)
        ctx.should_call_api(
            "send_msg",
            {
                "message_type": "group",
                "group_id": 456,
                "message": MessageSegment.reply(1) + MessageSegment.at(123) + MessageSegment.text("hello!"),
            },
        )
        ctx.should_call_api(
            "send_msg",
            {
                "message_type": "group",
                "group_id": 456,
                "message": MessageSegment.reply(2) + MessageSegment.at(123) + MessageSegment.text("world!"),
            },
        )
        ctx.should_call_api("delete_msg", {"message_id": 2})

    async with app.test_api() as ctx1:
        adapter = get_adapter(Adapter)
        _ = ctx1.create_bot(base=Bot, adapter=adapter)
        ctx1.should_call_api(
            "send_msg",
            {
                "message_type": "group",
                "group_id": 456,
                "message": Message("hello!"),
            },
        )
        target = Target("456", platform=adapter.get_name())
        await target.send("hello!")



from typing import Union

from nepattern.base import URL, INTEGER
from nonebot.adapters.qq.message import Ark as _Ark
from nonebot.adapters.qq.message import MessageSegment
from nonebot.adapters.qq.message import Embed as _Embed
from nonebot.adapters.qq.message import Emoji as _Emoji
from nepattern import MatchMode, BasePattern, UnionPattern
from nonebot.adapters.qq.message import Keyboard as _Keyboard
from nonebot.adapters.qq.message import Markdown as _Markdown
from nonebot.adapters.qq.message import Reference as _Reference
from nonebot.adapters.qq.message import Attachment as _Attachment
from nonebot.adapters.qq.message import MentionUser as _MentionUser
from nonebot.adapters.qq.message import MentionChannel as _MentionChannel
from nonebot.adapters.qq.message import LocalAttachment as _LocalAttachment
from nonebot.adapters.qq.message import MentionEveryone as _MentionEveryone

from nonebot_plugin_alconna.typings import SegmentPattern

Text = str
Ark = SegmentPattern("ark", _Ark, MessageSegment.ark)
Embed = SegmentPattern("embed", _Embed, MessageSegment.embed)
Emoji = SegmentPattern("emoji", _Emoji, MessageSegment.emoji)
Image = SegmentPattern("image", _Attachment, MessageSegment.image)
FileImage = SegmentPattern("file_image", _LocalAttachment, MessageSegment.file_image)
Audio = SegmentPattern("audio", _Attachment, MessageSegment.audio)
FileAudio = SegmentPattern("file_audio", _LocalAttachment, MessageSegment.file_audio)
Video = SegmentPattern("video", _Attachment, MessageSegment.video)
FileVideo = SegmentPattern("file_video", _LocalAttachment, MessageSegment.file_video)
File = SegmentPattern("file", _Attachment, MessageSegment.file)
FileFile = SegmentPattern("file_file", _LocalAttachment, MessageSegment.file_file)
Keyboard = SegmentPattern("keyboard", _Keyboard, MessageSegment.keyboard)
Markdown = SegmentPattern("markdown", _Markdown, MessageSegment.markdown)
MentionUser = SegmentPattern("mention_user", _MentionUser, MessageSegment.mention_user)
MentionChannel = SegmentPattern("mention_channel", _MentionChannel, MessageSegment.mention_channel)
MentionEveryone = SegmentPattern("mention_everyone", _MentionEveryone, MessageSegment.mention_everyone)
Reference = SegmentPattern("reference", _Reference, MessageSegment.reference)


ImgOrUrl = (
    UnionPattern(
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=str,
                converter=lambda _, x: x.data["url"],
                alias="img",
                addition_accepts=Image,
            ),
            URL,
        ]
    )
    @ "img_url"
)
"""
内置类型, 允许传入图片元素(Image)或者链接(URL)，返回链接
"""

MentionID = (
    UnionPattern[Union[str, _MentionUser]](
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=int,
                alias="MentionUser",
                addition_accepts=MentionUser,
                converter=lambda _, x: int(x.data["user_id"]),
            ),
            BasePattern(
                r"@(\d+)",
                mode=MatchMode.REGEX_CONVERT,
                origin=int,
                alias="@xxx",
                accepts=str,
                converter=lambda _, x: int(x[1]),
            ),
            INTEGER,
        ]
    )
    @ "mention_id"
)
"""
内置类型，允许传入提醒元素(Mention)或者'@xxxx'式样的字符串或者数字, 返回数字
"""



MAPPING = {
    "mirai2": "mirai",
    "BilibiliLive": "bilibili",
    "Console": "console",
    "Ding": "ding",
    "Discord": "discord",
    "Dodo": "dodo",
    "Feishu": "feishu",
    "Github": "github",
    "Kaiheila": "kook",
    "Minecraft": "minecraft",
    "ntchat": "ntchat",
    "OneBot V11": "onebot11",
    "OneBot V12": "onebot12",
    "QQ": "qq",
    "QQ Guild": "qqguild",
    "RedProtocol": "red",
    "Telegram": "telegram",
    "Satori": "satori",
    "Villa": "villa",
}



from pathlib import Path
from typing import Literal

from tarina import lang
from nonebot.utils import logger_wrapper

lang.load(Path(__file__).parent / "i18n")
ALCONNA_RESULT: Literal["_alc_result"] = "_alc_result"
ALCONNA_EXEC_RESULT: Literal["_alc_exec_result"] = "_alc_exec_result"
ALCONNA_ARG_KEY: Literal["_alc_arg_{key}"] = "_alc_arg_{key}"
ALCONNA_EXTENSION: Literal["_alc_extension"] = "_alc_extension"

log = logger_wrapper("Plugin-Alconna")



import random
from pathlib import Path
from base64 import b64decode
from typing import TYPE_CHECKING, List, Type, Union, Literal, Optional, overload

from yarl import URL
from nonebot import get_bots
from nonebot.typing import T_State
from nonebot import get_bot as _get_bot
from nonebot.exception import ActionFailed
from nonebot.internal.driver.model import Request
from nonebot.internal.adapter import Bot, Event, Adapter

from .segment import Image


async def image_fetch(event: Event, bot: Bot, state: T_State, img: Image, **kwargs):
    adapter_name = bot.adapter.get_name()
    if adapter_name == "RedProtocol":
        origin = img.origin
        if TYPE_CHECKING:
            from nonebot.adapters.red.bot import Bot
            from nonebot.adapters.red.message import MediaMessageSegment

            assert isinstance(bot, Bot)
            assert isinstance(origin, MediaMessageSegment)

        return await origin.download(bot)

    if img.url:  # mirai2, qqguild, kook, villa, minecraft, ding
        req = Request("GET", img.url, **kwargs)
        resp = await bot.adapter.request(req)
        return resp.content
    if not img.id:
        return None
    if adapter_name == "OneBot V11":
        if TYPE_CHECKING:
            from nonebot.adapters.onebot.v11.bot import Bot

            assert isinstance(bot, Bot)
        url = (await bot.get_image(file=img.id))["data"]["url"]
        req = Request("GET", url, **kwargs)
        resp = await bot.adapter.request(req)
        return resp.content
    if adapter_name == "OneBot V12":
        if TYPE_CHECKING:
            from nonebot.adapters.onebot.v12.bot import Bot

            assert isinstance(bot, Bot)
        resp = (await bot.get_file(type="data", file_id=img.id))["data"]
        return b64decode(resp) if isinstance(resp, str) else bytes(resp)
    if adapter_name == "mirai2":
        url = f"https://gchat.qpic.cn/gchatpic_new/0/0-0-" f"{img.id.replace('-', '').upper()}/0"
        req = Request("GET", url, **kwargs)
        resp = await bot.adapter.request(req)
        return resp.content
    if adapter_name == "Telegram":
        if TYPE_CHECKING:
            from nonebot.adapters.telegram.bot import Bot

            assert isinstance(bot, Bot)
        res = await bot.get_file(file_id=img.id)
        if not res.file_path:
            raise ActionFailed("Telegram", "get file failed")
        if (p := Path(res.file_path)).exists():  # telegram api local mode
            return p.read_bytes()
        url = URL(bot.bot_config.api_server) / "file" / f"bot{bot.bot_config.token}" / res.file_path
        req = Request("GET", url, **kwargs)
        resp = await bot.adapter.request(req)
        return resp.content
    if adapter_name == "Feishu":
        if TYPE_CHECKING:
            from nonebot.adapters.feishu.bot import Bot
            from nonebot.adapters.feishu.event import MessageEvent

            assert isinstance(bot, Bot)
            assert isinstance(event, MessageEvent)
        return await bot.get_msg_resource(message_id=event.message_id, file_key=img.id, type_="image")
    if adapter_name == "ntchat":
        raise NotImplementedError("ntchat image fetch not implemented")


@overload
def get_bot(*, index: int) -> Bot: ...


@overload
def get_bot(*, rand: Literal[True]) -> Bot: ...


@overload
def get_bot(*, bot_id: str) -> Bot: ...


@overload
def get_bot(*, adapter: Union[Type[Adapter], str]) -> List[Bot]: ...


@overload
def get_bot(*, adapter: Union[Type[Adapter], str], index: int) -> Bot: ...


@overload
def get_bot(*, adapter: Union[Type[Adapter], str], rand: Literal[True]) -> Bot: ...


@overload
def get_bot(*, adapter: Union[Type[Adapter], str], bot_id: str) -> Bot: ...


def get_bot(
    *,
    adapter: Union[Type[Adapter], str, None] = None,
    bot_id: Optional[str] = None,
    index: Optional[int] = None,
    rand: bool = False,
) -> Union[List[Bot], Bot]:
    if not adapter:
        if rand:
            return random.choice(list(get_bots().values()))
        if index is not None:
            return list(get_bots().values())[index]
        return _get_bot(bot_id)
    bots = []
    for bot in get_bots().values():
        _adapter = bot.adapter
        if isinstance(adapter, str):
            if _adapter.get_name() == adapter:
                bots.append(bot)
        elif isinstance(_adapter, adapter):
            bots.append(bot)
    if not bot_id:
        if rand:
            return random.choice(bots)
        if index is not None:
            return bots[index]
        return bots
    return next(bot for bot in bots if bot.self_id == bot_id)



from typing import Union

from nepattern.base import URL, INTEGER
from nepattern import MatchMode, BasePattern, UnionPattern
from nonebot.adapters.dodo.message import (
    CardSegment,
    FileSegment,
    AtAllSegment,
    ShareSegment,
    VideoSegment,
    AtRoleSegment,
    AtUserSegment,
    MessageSegment,
    PictureSegment,
    ReferenceSegment,
    ChannelLinkSegment,
)

from nonebot_plugin_alconna.typings import SegmentPattern

Text = str
At = AtUser = SegmentPattern("at_user", AtUserSegment, MessageSegment.at_user)
ChannelLink = SegmentPattern("channel_link", ChannelLinkSegment, MessageSegment.channel_link)
Reference = SegmentPattern("reference", ReferenceSegment, MessageSegment.reference)
Image = Picture = SegmentPattern("picture", PictureSegment, MessageSegment.picture)
Video = SegmentPattern("video", VideoSegment, MessageSegment.video)
Card = SegmentPattern("card", CardSegment, MessageSegment.card)
AtRole = SegmentPattern("at_role", AtRoleSegment, AtRoleSegment)
AtAll = SegmentPattern("at_all", AtAllSegment, AtAllSegment)
Share = SegmentPattern("share", ShareSegment, ShareSegment)
File = SegmentPattern("file", FileSegment, FileSegment)

ImgOrUrl = (
    UnionPattern(
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=str,
                converter=lambda _, x: x.data["picture"].url,
                alias="img",
                addition_accepts=Image,
            ),
            URL,
        ]
    )
    @ "img_url"
)
"""
内置类型, 允许传入图片元素(Image)或者链接(URL)，返回链接
"""

AtID = (
    UnionPattern[Union[str, AtUserSegment]](
        [
            BasePattern(
                mode=MatchMode.TYPE_CONVERT,
                origin=int,
                alias="AtUser",
                addition_accepts=AtUser,
                converter=lambda _, x: int(x.data["dodo_id"]),
            ),
            BasePattern(
                r"@(\d+)",
                mode=MatchMode.REGEX_CONVERT,
                origin=int,
                alias="@xxx",
                accepts=str,
                converter=lambda _, x: int(x[1]),
            ),
            INTEGER,
        ]
    )
    @ "at_id"
)
"""
内置类型，允许传入提醒元素(AtUser)或者'@xxxx'式样的字符串或者数字, 返回数字
"""



import pytest
from nonebug import App
from nonebot import get_adapter
from nonebot.adapters.satori.models import User
from arclet.alconna import Args, Alconna, CommandMeta
from nonebot.adapters.satori import Bot, Adapter, Message

from tests.fake import fake_message_event_satori


@pytest.mark.asyncio()
async def test_ctx(app: App):
    from nonebot_plugin_alconna import Text, on_alconna

    test_cmd = on_alconna(Alconna("test", Args["userid", str], meta=CommandMeta(context_style="parentheses")))

    @test_cmd.handle()
    async def tt_h(userid: str, ctx: dict):
        assert ctx["event"].get_user_id() == userid
        await test_cmd.send(Text("ok\n") + userid)

    async with app.test_matcher(test_cmd) as ctx:
        adapter = get_adapter(Adapter)
        bot = ctx.create_bot(base=Bot, adapter=adapter, platform="satori", info=None)
        msg = "test $(event.get_user_id())"
        event = fake_message_event_satori(message=msg, id=123, user=User(id="456", name="test"))
        ctx.receive_event(bot, event)
        ctx.should_call_send(event, Message("ok\n456"))



from __future__ import annotations

import re
import asyncio
import functools
import importlib as imp
from weakref import finalize
from dataclasses import dataclass
from typing_extensions import Self
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Union, Generic, Literal, TypeVar

from tarina import lang
from nonebot.typing import T_State
from arclet.alconna import Alconna, Arparma
from nonebot.compat import PydanticUndefined
from nonebot.adapters import Bot, Event, Message

from .uniseg import UniMessage, FallbackMessage

OutputType = Literal["help", "shortcut", "completion"]
TM = TypeVar("TM", bound=Union[str, Message, UniMessage])
TE = TypeVar("TE", bound=Event)

if TYPE_CHECKING:
    from .rule import AlconnaRule


@dataclass
class Interface(Generic[TE]):
    event: TE
    state: T_State
    name: str
    annotation: Any
    default: Any


class Extension(metaclass=ABCMeta):
    _overrides: dict[str, bool]

    def __init_subclass__(cls, **kwargs):
        cls._overrides = {
            "output_converter": cls.output_converter != Extension.output_converter,
            "send_wrapper": cls.send_wrapper != Extension.send_wrapper,
            "receive_wrapper": cls.receive_wrapper != Extension.receive_wrapper,
            "permission_check": cls.permission_check != Extension.permission_check,
            "context_provider": cls.context_provider != Extension.context_provider,
            "parse_wrapper": cls.parse_wrapper != Extension.parse_wrapper,
            "catch": cls.catch != Extension.catch and cls.before_catch != Extension.before_catch,
        }

    @property
    @abstractmethod
    def priority(self) -> int:
        """插件优先级。"""
        raise NotImplementedError

    @property
    @abstractmethod
    def id(self) -> str:
        """插件 ID。"""
        raise NotImplementedError

    @property
    def namespace(self) -> str:
        """插件所属命名空间"""
        return ""

    def validate(self, bot: Bot, event: Event) -> bool:
        return True

    async def output_converter(self, output_type: OutputType, content: str) -> Message | UniMessage:
        """依据输出信息的类型，将字符串转换为消息对象以便发送。"""
        return FallbackMessage(content)

    async def message_provider(
        self, event: Event, state: T_State, bot: Bot, use_origin: bool = False
    ) -> Message | UniMessage | None:
        """提供消息对象以便 Alconna 进行处理。"""
        if event.get_type() != "message":
            return None
        msg: Message = event.get_message()
        if use_origin:
            try:
                msg: Message = getattr(event, "original_message", msg)  # type: ignore
            except (NotImplementedError, ValueError):
                return None
        return msg

    async def receive_wrapper(self, bot: Bot, event: Event, command: Alconna, receive: TM) -> TM:
        """接收消息后的钩子函数。"""
        return receive

    async def permission_check(self, bot: Bot, event: Event, command: Alconna) -> bool:
        """命令首次解析并确认头部匹配（即确认选择响应）时对发送者的权限判断"""
        return True

    async def context_provider(self, ctx: dict[str, Any], event: Event, bot: Bot, state: T_State) -> dict[str, Any]:
        """提供上下文以便 Alconna 进行处理。"""
        return ctx

    async def parse_wrapper(self, bot: Bot, state: T_State, event: Event, res: Arparma) -> None:
        """解析消息后的钩子函数。"""
        ...

    async def send_wrapper(self, bot: Bot, event: Event, send: TM) -> TM:
        """发送消息前的钩子函数。"""
        return send

    def before_catch(self, name: str, annotation: Any, default: Any) -> bool:
        """依赖注入的绑定确认函数。"""
        ...

    async def catch(self, interface: Interface) -> Any:
        """自定义依赖注入处理函数。"""
        ...

    def post_init(self, alc: Alconna) -> None:
        """Alconna 初始化后的钩子函数。"""
        ...


class DefaultExtension(Extension):
    @property
    def priority(self) -> int:
        return 16

    @property
    def id(self) -> str:
        return "!default"


_callbacks = set()


class ExtensionExecutor:
    globals: list[type[Extension] | Extension] = [DefaultExtension()]
    _rule: AlconnaRule

    def __init__(
        self,
        rule: AlconnaRule,
        extensions: list[type[Extension] | Extension] | None = None,
        excludes: list[str | type[Extension]] | None = None,
    ):
        self.extensions: list[Extension] = []
        for ext in self.globals:
            if isinstance(ext, type):
                self.extensions.append(ext())
            else:
                self.extensions.append(ext)
        for ext in extensions or []:
            if isinstance(ext, type):
                self.extensions.append(ext())
            else:
                self.extensions.append(ext)
        for exl in excludes or []:
            if isinstance(exl, str) and exl.startswith("!"):
                raise ValueError(lang.require("nbp-alc", "error.extension_forbid_exclude"))
        self._excludes = set(excludes or [])
        self.extensions = [
            ext
            for ext in self.extensions
            if ext.id not in self._excludes
            and ext.__class__ not in self._excludes
            and (not (ns := ext.namespace) or ns == rule.command.namespace)
        ]
        self.context: list[Extension] = []
        self._rule = rule

        _callbacks.add(self._callback)

        finalize(self, _callbacks.remove, self._callback)

    def _callback(self, *append_global_ext: type[Extension] | Extension):
        for _ext in append_global_ext:
            if isinstance(_ext, type):
                _ext = _ext()
            if _ext.id in self._excludes or _ext.__class__ in self._excludes:
                continue
            if (ns := _ext.namespace) and ns != self._rule.command.namespace:
                continue
            self.extensions.append(_ext)
            _ext.post_init(self._rule.command)

    def __exit__(self, exc_type, exc_value, traceback):
        self.context.clear()

    def select(self, bot: Bot, event: Event) -> Self:
        self.context = [ext for ext in self.extensions if ext.validate(bot, event)]
        self.context.sort(key=lambda ext: ext.priority)
        return self

    async def output_converter(self, output_type: OutputType, content: str) -> Message | UniMessage:
        exc = None
        for ext in self.context:
            if not ext._overrides["output_converter"]:
                continue
            try:
                return await ext.output_converter(output_type, content)
            except Exception as e:
                exc = e
        if not exc:
            return FallbackMessage()
        raise exc  # type: ignore

    async def message_provider(
        self, event: Event, state: T_State, bot: Bot, use_origin: bool = False
    ) -> Message | UniMessage | None:
        exc = None
        for ext in self.context:
            try:
                if (msg := await ext.message_provider(event, state, bot, use_origin)) is not None:
                    return msg
            except Exception as e:
                exc = e
        if exc is not None:
            raise exc

        return None

    async def receive_wrapper(self, bot: Bot, event: Event, receive: TM) -> TM:
        res = receive
        for ext in self.context:
            if ext._overrides["receive_wrapper"]:
                res = await ext.receive_wrapper(bot, event, self._rule.command, res)
        return res

    async def permission_check(self, bot: Bot, event: Event) -> bool:
        for ext in self.context:
            if ext._overrides["permission_check"]:
                if await ext.permission_check(bot, event, self._rule.command) is False:
                    return False
                continue
        return True

    async def context_provider(self, event: Event, bot: Bot, state: T_State) -> dict[str, Any]:
        ctx = {}
        for ext in self.context:
            if ext._overrides["context_provider"]:
                ctx = await ext.context_provider(ctx, event, bot, state)
        ctx["event"] = event
        # ctx["bot"] = bot
        return ctx

    async def parse_wrapper(self, bot: Bot, state: T_State, event: Event, res: Arparma) -> None:
        await asyncio.gather(
            *(ext.parse_wrapper(bot, state, event, res) for ext in self.context if ext._overrides["parse_wrapper"])
        )

    async def send_wrapper(self, bot: Bot, event: Event, send: TM) -> TM:
        res = send
        for ext in self.context:
            if ext._overrides["send_wrapper"]:
                res = await ext.send_wrapper(bot, event, res)
        return res

    def before_catch(self, name: str, annotation: Any, default: Any) -> bool:
        for ext in self.extensions:
            if ext._overrides["catch"]:
                if ext.before_catch(name, annotation, default):
                    return True
                continue
        return False

    async def catch(self, event: Event, state: T_State, name: str, annotation: Any, default: Any):
        for ext in self.extensions:
            if ext._overrides["catch"]:
                res = await ext.catch(Interface(event, state, name, annotation, default))
                if res is None:
                    continue
                return res
        return PydanticUndefined

    def post_init(self) -> None:
        for ext in self.extensions:
            ext.post_init(self._rule.command)


def add_global_extension(*ext: type[Extension] | Extension) -> None:
    ExtensionExecutor.globals.extend(ext)
    for callback in _callbacks:
        callback(*ext)


pattern = re.compile(r"(?P<module>[\w.]+)\s*" r"(:\s*(?P<attr>[\w.]+)\s*)?" r"((?P<extras>\[.*\])\s*)?$")


def load_from_path(path: str) -> None:
    if path.startswith("~."):
        path = f"nonebot_plugin_alconna{path[1:]}"
    elif path.startswith("~"):
        path = f"nonebot_plugin_alconna.{path[1:]}"
    match = pattern.match(path)
    if not match:
        raise ValueError(lang.require("nbp-alc", "error.extension_path_invalid").format(path=path))
    module = imp.import_module(match.group("module"))
    attrs = filter(None, (match.group("attr") or "__extension__").split("."))
    ext = functools.reduce(getattr, attrs, module)
    if isinstance(ext, type) and issubclass(ext, Extension):
        add_global_extension(ext)
    elif isinstance(ext, Extension):
        add_global_extension(ext)
    else:
        raise TypeError(lang.require("nbp-alc", "error.extension_path_load").format(path=path))
